\ EXCEPTIO.F            Exception Support               by Tom Zimmer

cr .( Loading Windows Exception...)

only forth also definitions 

INTERNAL

0 value ?exceptioning                   \ detect recursive call to .exception

: .exception-code ( -- )                \ get the last exception
                exception@
case
   STATUS_WAIT_0                   of ." WAIT_0"                        endof
   STATUS_ABANDONED_WAIT_0         of ." ABANDONED_WAIT_0"              endof
   STATUS_USER_APC                 of ." USER_APC"                      endof
   STATUS_TIMEOUT                  of ." TIMEOUT"                       endof
   STATUS_PENDING                  of ." PENDING"                       endof
   STATUS_DATATYPE_MISALIGNMENT    of ." DATATYPE_MISALIGNMENT"         endof
   STATUS_BREAKPOINT               of ." BREAKPOINT"                    endof
   STATUS_SINGLE_STEP              of ." SINGLE_STEP"                   endof
   STATUS_ACCESS_VIOLATION         of ." ACCESS_VIOLATION"              endof
   STATUS_IN_PAGE_ERROR            of ." IN_PAGE_ERROR"                 endof
   STATUS_NO_MEMORY                of ." NO_MEMORY"                     endof
   STATUS_ILLEGAL_INSTRUCTION      of ." ILLEGAL_INSTRUCTION"           endof
   STATUS_NONCONTINUABLE_EXCEPTION of ." NONCONTINUABLE_EXCEPTION"      endof
   STATUS_INVALID_DISPOSITION      of ." INVALID_DISPOSITION"           endof
   STATUS_ARRAY_BOUNDS_EXCEEDED    of ." ARRAY_BOUNDS_EXCEEDED"         endof
   STATUS_FLOAT_DENORMAL_OPERAND   of ." FLOAT_DENORMAL_OPERAND"        endof
   STATUS_FLOAT_DIVIDE_BY_ZERO     of ." FLOAT_DIVIDE_BY_ZERO"          endof
   STATUS_FLOAT_INEXACT_RESULT     of ." FLOAT_INEXACT_RESULT"          endof
   STATUS_FLOAT_INVALID_OPERATION  of ." FLOAT_INVALID_OPERATION"       endof
   STATUS_FLOAT_OVERFLOW           of ." FLOAT_OVERFLOW"                endof
   STATUS_FLOAT_STACK_CHECK        of ." FLOAT_STACK_CHECK"             endof
   STATUS_FLOAT_UNDERFLOW          of ." FLOAT_UNDERFLOW"               endof
   STATUS_INTEGER_DIVIDE_BY_ZERO   of ." INTEGER_DIVIDE_BY_ZERO"        endof
   STATUS_INTEGER_OVERFLOW         of ." INTEGER_OVERFLOW"              endof
   STATUS_PRIVILEGED_INSTRUCTION   of ." PRIVILEGED_INSTRUCTION"        endof
   STATUS_STACK_OVERFLOW           of ." STACK_OVERFLOW"                endof
   STATUS_CONTROL_C_EXIT           of ." CONTROL_C_EXIT"                endof
                                      dup ." 0x" 1 h.r ."  "
endcase ;

: context-buffer ( -- a1 )              \ return the address of the
                &exrec @ abs>rel ;  \ exception handlers context buffer

: ex-rstack     ( -- a1 )               \ exception return stack base
                &exrstk ;

0       nostack1
        cell  + \ field+ ContextFlags
        cell  + \ field+ Dr0
        cell  + \ field+ Dr1
        cell  + \ field+ Dr2
        cell  + \ field+ Dr3
        cell  + \ field+ Dr6
        cell  + \ field+ Dr7
        0     + \ field+ FloatSaveArea
        cell  + \ field+ FloatContro
        cell  + \ field+ FloatStatus
        cell  + \ field+ FloatTag
        cell  + \ field+ FloatErrorOffset
        cell  + \ field+ FloatErrorSelector
        cell  + \ field+ FloatDataOffset
        80    + \ field+ FloatDataSelector
        cell  + \ field+ FloatRegisterArea
        cell  + \ field+ FloatCr0NpxState
        cell  + \ field+ SegGs
        cell  + \ field+ SegFs
        cell  + \ field+ SegEs
        cell  + \ field+ SegDs
        cell  field+ Edi
        cell  field+ Esi
        cell  field+ Ebx
        cell  field+ Edx
        cell  field+ Ecx
        cell  field+ Eax
        cell  field+ Ebp
        cell  field+ Eip
        cell  field+ SegCs
        cell  field+ EFlags
        cell  field+ Esp
        cell  field+ SegSs
            constant context-length

EXTERNAL

: clear-exception ( -- )                \ clear any previous exception
                &except off ;

: .exception    ( -- )
                exception@ 0=
        if      ."  * No Exception has occured * "
        else    forth-io
                cr ." EXCEPTION: " >bold .exception-code >norm
                ?exceptioning
                if      cr ." Recursive Exception canceled"
                        exit            \ terminate .exception
                then
                true to ?exceptioning
                context-buffer
                cr ."   Registers:"
                cr ."     Eax:"          dup Eax             @ 14 col 8 h.r ." h"
                cr ."     Ebx:"          dup Ebx             @ 14 col 8 h.r ." h   top of data stack"
                cr ."     Edx:"          dup Edx             @ 14 col 8 h.r ." h"
                cr ."     Ecx:"          dup Ecx             @ 14 col 8 h.r ." h"
                cr ."     Esi: IP"       dup Esi  @ abs>rel    14 col 8 h.r ." h   image relative"
                cr ."     Edi: IMAGE"    dup Edi             @ 14 col 8 h.r ." h   Forth's base address"
                cr ."     Esp: SP@"      dup Esp  @ abs>rel    14 col 8 h.r ." h   image relative"
                cr ."     Ebp: RP@"      dup Ebp  @ abs>rel    14 col 8 h.r ." h   image relative"
                cr ."     Eip: PC"       dup Eip  @ abs>rel    14 col 8 h.r ." h   image relative"
        [ also bug ]    ex-rstack dup 4 cells+ #dbg-rstack
        [ previous ]
          >bold cr ."   At forth word: " dup Eip  @ abs>rel    20 col ?name .NAME >norm
\               cr ."     SegCs:"        dup SegCs           @ 14 col 8 h.r ." h"
\               cr ."     EFlags:"       dup EFlags          @ 14 col 8 h.r ." h"
\               cr ."     SegSs:"        dup SegSs           @ 14 col 8 h.r ." h"
\               cr ."     ContextFlags:" dup ContextFlags    @ 14 col 8 h.r ." h"
\               cr ."     Dr0:"          dup Dr0             @ 14 col 8 h.r ." h"
\               cr ."     Dr1:"          dup Dr1             @ 14 col 8 h.r ." h"
\               cr ."     Dr2:"          dup Dr2             @ 14 col 8 h.r ." h"
\               cr ."     Dr3:"          dup Dr3             @ 14 col 8 h.r ." h"
\               cr ."     Dr6:"          dup Dr6             @ 14 col 8 h.r ." h"
\               cr ."     Dr7:"          dup Dr7             @ 14 col 8 h.r ." h"
\               cr ."     SegGs:"        dup SegGs           @ 14 col 8 h.r ." h"
\               cr ."     SegFs:"        dup SegFs           @ 14 col 8 h.r ." h"
\               cr ."     SegEs:"        dup SegEs           @ 14 col 8 h.r ." h"
\               cr ."     SegDs:"        dup SegDs           @ 14 col 8 h.r ." h"
                drop
                false to ?exceptioning
        then    ;


MODULE

only forth also definitions


