\ FKERNEL.FTH  - WIN32FORTH KERNEL
\ Andrew McKewan
\ March 1994
\ Given to Tom Zimmer 05/13/94, assemblable with TASM32
\ Metacompiler version 11/95 Andrew McKewan
\ Separated heads version started December 19th, 1995 tjz
\ Added User variables for multi-tasking August 29th, 1996 bee/tjz

DECIMAL

\        sp     equ     <esp>   \ Stack Pointer for Forth, the hardware stack
\        rp     equ     <ebp>   \ Return Pointer, Forth's subroutine stack
\        ip     equ     <esi>   \ "absolute" Instruction Pointer for Forth 
\        bp     equ     <edi>   \ "absolute" base of Forth image
\        up     equ     <edx>   \ "absolute" user pointer 
\        tos    equ     <ebx>   \ Top of stack is in EBX

26350037 EQU MAGIC#         \ Magic # to indentify a dictionary image
260      EQU maxbuffer      \ Size of any string buffer, MUST match the
                            \ size of a windows maximum path string,
                            \ which is 260 bytes. ** DON'T CHANGE THIS **
16       EQU numvocs        \ Maximum number of vocabularies in search order

 -1 EQU THROW_ABORT
 -2 EQU THROW_ABORTQ
 -3 EQU THROW_STACK_OVERFLOW
 -4 EQU THROW_STACK_UNDERFLOW
-13 EQU THROW_UNDEFINED


\ -------------------- Inner Interpreter --------------------

MACRO EXEC      ( -- )                  \ execute RELATIVE cfa in eax
                mov     ecx, 0 [eax] [edi]
                add     ecx, edi
                jmp     ecx
                END-MACRO

MACRO NEXT      ( -- )                  \ Inner interpreter
                RESOLVE-OFA             \ save size of code definition
                mov     eax, 0 [esi]    \ faster version of NEXT, uses 9 more
                add     esi, # 4        \ bytes per occurance
                exec
                END-MACRO

MACRO 16:       ( -- )                  \ compile 16-bit prefix opcode
                A;
                102 C,-T
                END-MACRO

\ -------------------- Boot-Up Literals --------------------

HERE: ORIG        ORIGIN ,        \ My physical starting addres
HERE: MAGIC       MAGIC# ,        \ magic number (make sure I am ME)
HERE: ASIZE       APPSIZE ,       \ desired dictionary size in bytes
HERE: AACTUAL     0 ,             \ actual dictionary length in bytes
HERE: SSIZE       SYSSIZE ,       \ desired head space size in bytes
HERE: SACTUAL     0 ,             \ actual head length in bytes
HERE: SSEPARATE   SSEPARATION ,   \ code to head separation (multiple of 64k)
HERE: ENTRY       0 ,             \ cold start entry point
HERE: TABLE       0 ,             \ pointer to xcall table entries
HERE: HINST       0 ,             \ applications instance handle
HERE: CMDLEN      0 ,             \ length  of command line
HERE: CMDADR      0 ,             \ address of command line
HERE: HCON        0 ,             \ applications console handle
HERE: EXCEPT      0 ,             \ exception variable, defaults to none
HERE: EXREC       0 ,             \ a pointer to the expection record
HERE: EXRSTK      0 , 0 , 0 , 0 , \ room for top return stack entries,
HERE: UNFORTH     0 ,             \ un-initialize forth
HERE: DLLINIT     0 ,             \ dll initialization function
HERE: DLLNTRY     0 ,             \ function entry for all DLL functions
HERE: MSGNTRY     0 ,             \ message callbacks for key and key?
HERE: WINMSGNTRY  0 ,             \ windows message callbacks


\ -------------------- Runtime Routines --------------------

LABEL DOCOL     ( -- )                  \ runtime for colon definitions
optimize? 0= #if
                sub     ebp, # 4
                mov     0 [ebp], esi
                lea     esi, 4 [eax] [edi]
                next    c;
#else           \ more efficient version of DOCOL with a modified version
                \ of NEXT integrated into its definition
                mov     -4 [ebp], esi
                lea     esi, 8 [eax] [edi]
                sub     ebp, # 4
                mov     eax, -4 [esi]      
                mov     ecx, 0 [eax] [edi]
                add     ecx, edi
                jmp     ecx
                c;
#then

LABEL DODOES    ( -- )                  \ runtime for DOES>
                sub     ebp, # 4
                mov     0 [ebp], esi
                pop     esi
                push    ebx
                lea     ebx, 4 [eax]
                next    c;

LABEL DOUSER    ( -- a1 )               \ runtime for USER variables
                push    ebx
                mov     ebx, 4 [eax] [edi]      \ get offset
                add     ebx, edx                \ add absolute user base
                sub     ebx, edi                \ remove absolute forth base
                next    c;

LABEL DOCON     ( -- n1 )               \ runtime for constants
                push    ebx
                mov     ebx, 4 [eax] [edi]
                next    c;

LABEL DOVAR     ( -- a1 )               \ runtime for CREATE and VARIABLE
                push    ebx
                lea     ebx, 4 [eax]
                next    c;

LABEL DODEFER   ( -- )                  \ runtime for DEFER
                mov     eax, 4 [eax] [edi]
                exec    c;

LABEL DOVALUE   ( -- n1 )               \ runtime for VALUE fetch
                push    ebx
                mov     ebx, 4 [eax] [edi]
                next    c;

LABEL DOVALUE!  ( n1 -- )               \ runtime for VALUE store
                mov     -4 [eax] [edi], ebx
                pop     ebx
                next    c;

LABEL DOVALUE+! ( n1 -- )               \ runtime for VALUE increment
                add     -8 [eax] [edi], ebx
                pop     ebx
                next    c;

LABEL DO2VALUE  ( d1 -- )               \ runtime for 2VALUE fetch
                push    ebx
                mov     ebx, 4 [eax] [edi]
                mov     eax, 4 [ebx] [edi]
                mov     ebx, 0 [ebx] [edi]
                push    eax
                next    c;

\ ----------------- LOADFILE linkage -----------------

HERE: KERNEL                    \ just before the first header
        0 ,                     \ link field
        ," FKERNEL.F"           \ define this filename
        MAXBUFFER ALLOT


CREATE LOADFILE                 \ head of the linked list of files loaded
       KERNEL ,                 \ points initially to this file

MAXBUFFER CONSTANT MAXSTRING    \ maximum length of a counted string
MAXBUFFER CONSTANT MAX-PATH     \ maximum length of a filename buffer
      255 CONSTANT MAXCOUNTED   \ maximum length of contents of a counted string

ORIGIN CONSTANT ^IMAGE          \ The pointer to the start of the Image

\ NEXT-SEQ returns the address of a counted string that is the code
\ compiled by NEXT. Used by the disassembler.

CREATE NEXT-SEQ
        0 OFA-H !               \ disable OFA resolution
      HERE 0 C,
        INIT-ASSEMBLER
        NEXT
        END-CODE
      HERE OVER - 1- SWAP C!-T

VARIABLE DEFER-LIST             \ The head of a linked-list of deferred words

\ -------------------- Primatives --------------------

VOCABULARY FORTH

\ Define the Forth base image pointer constants. These constants point to
\ fields at the beginning of Forth that hold various information that must
\ be sent to or received from the wrapper.

          ORIG               CONSTANT &ORIGIN
          MAGIC              CONSTANT &MAGIC
          ASIZE              CONSTANT &APP-SIZE
          AACTUAL            CONSTANT &APP-ACTUAL
          SSIZE              CONSTANT &SYS-SIZE
          SACTUAL            CONSTANT &SYS-ACTUAL
          SSEPARATE          CONSTANT &SYS-SEPARATION
          ENTRY              CONSTANT &ENTRY
          TABLE              CONSTANT &TABLE
          HINST              CONSTANT &HINST
          CMDLEN             CONSTANT &CMDLEN
          CMDADR             CONSTANT &CMDADR
          HCON               CONSTANT &HCON
          EXCEPT             CONSTANT &EXCEPT
          EXREC              CONSTANT &EXREC
          EXRSTK             CONSTANT &EXRSTK
          UNFORTH            CONSTANT &UNLOAD-FORTH
          DLLINIT            CONSTANT &DLLINIT
          DLLNTRY            CONSTANT &DLL-CALLBACK
          MSGNTRY            CONSTANT &MESSAGE-CALLBACK
          WINMSGNTRY         CONSTANT &WINMESSAGE-CALLBACK

\ Define pointer constants for the basic Forth "runtime" routines.  A pointer
\ to one of these small assembly language routines is installed at the start
\ of every Forth definitions created.  They can actually be thought of as the
\ only form of Forth data "type"ing.  That is it is possible to look at the
\ contents of the CFA of any Forth definition, evaluate what is there, and
\ determine what type of definition it is.

ASSEMBLER DOCOL         META CONSTANT DOCOL
ASSEMBLER DOCON         META CONSTANT DOCON
ASSEMBLER DODOES        META CONSTANT DODOES
ASSEMBLER DOVAR         META CONSTANT DOVAR
ASSEMBLER DODEFER       META CONSTANT DODEFER
ASSEMBLER DO2VALUE      META CONSTANT DO2VALUE
ASSEMBLER DOVALUE       META CONSTANT DOVALUE
ASSEMBLER DOVALUE!      META CONSTANT DOVALUE!
ASSEMBLER DOVALUE+!     META CONSTANT DOVALUE+!
ASSEMBLER DOUSER        META CONSTANT DOUSER

NCODE _EXIT     ( -- )          \ exit the current Forth definition
                mov     esi, 0 [ebp]
                add     ebp, # 4
                next    c;

NCODE UNNEST    ( -- )          \ exit the current Forth definition
                mov     esi, 0 [ebp]
                add     ebp, # 4
                next    c;

CODE LIT        ( -- n )        \ push the literal value following LIT in the
                                \ dictionary onto the data stack
optimize? 0= #if
                push    ebx
                mov     ebx, 0 [esi]
                add     esi, # 4
                next    c;
#else           \ more efficient version of LIT with a modified version
                \ of NEXT integrated into its definition
                mov     eax, 4 [esi]
                push    ebx
                mov     ebx, 0 [esi]
                add     esi, # 8
                mov     ecx, 0 [eax] [edi]
                add     ecx, edi
                jmp     ecx
                c;
#then

NCODE EXECUTE   ( xt -- )       \ go perform the Forth function whose address
                                \ is "xt"
                mov     eax, ebx
                pop     ebx
                exec    c;

NCODE PERFORM  ( a1 -- )        \ go perform the Forth function whose address
                                \ is contained in address "a1"
                mov     eax, 0 [ebx] [edi]
                pop     ebx
                exec    c;

CODE NOOP       ( -- )          \ the Forth do-nothing function
                next    c;


\ -------------------- Branching & Looping --------------------

NCODE BRANCH    ( -- )          \ "runtime" for branch always
                add     esi, 0 [esi]
                next    c;

NCODE ?BRANCH   ( f1 -- )       \ "runtime" for branch on f1=FALSE
                test    ebx, ebx
                pop     ebx
                je      @@1
                add     esi, # 4
                jmp     short @@2
@@1:            add     esi, 0 [esi]
@@2:            next    c;

NCODE _BEGIN    ( -- )          \ "runtime" marker for the decompiler, a noop
                next    c;

NCODE _THEN     ( -- )          \ "runtime" marker for the decompiler, a noop
                next    c;

NCODE _UNTIL    ( f1 -- )       \ "runtime" if f1=FALSE branch to after BEGIN
                test    ebx, ebx
                pop     ebx
                jne     short @@1
                add     esi, 0 [esi]
                jmp     short @@2
@@1:            add     esi, # 4
@@2:            next    c;

NCODE _AGAIN    ( -- )          \ "runtime" branch back to after BEGIN
                add     esi, 0 [esi]
                next    c;

NCODE _WHILE    ( f1 -- )       \ "runtime" if f1=FALSE branch to after REPEAT
                test    ebx, ebx
                pop     ebx
                je      short @@1
                add     esi, # 4
                jmp     short @@2
@@1:            add     esi, 0 [esi]
@@2:            next    c;

NCODE _REPEAT   ( -- )          \ "runtime" branch back to after BEGIN
                add     esi, 0 [esi]
                next    c;

NCODE _CASE     ( -- )          \ "runtime" marker for the decompiler, a noop
                next    c;

NCODE _OF       ( n1 n2 -- [n1] ) \ "runtime"
                                  \ if n1<>n2 branch to after ENDOF, return n1
                                  \ else continue and don't return n1
                pop     eax
                cmp     ebx, eax
                je      short @@1
                mov     ebx, eax
                add     esi, 0 [esi]
                jmp     short @@2
@@1:            pop     ebx
                add     esi, # 4
@@2:            next    c;

NCODE _ENDOF    ( -- )          \ "runtime" branch to after ENDCASE
                add     esi, 0 [esi]
                next    c;

NCODE _ENDCASE  ( n1 -- )       \ "runtime" discard n1 and continue
                pop     ebx
                next    c;

NCODE (LOOP)    ( -- )          \ "runtime" bump count and branch to after
                                \ DO if loop count not complete
                inc     dword ptr 0 [ebp]
                jo      short @@1
                add     esi, 0 [esi]
                jmp     short @@2
@@1:            add     esi, # 4
                add     ebp, # 12
@@2:            next    c;

NCODE (+LOOP)   ( n1 -- )       \ "runtime" bump count by n1 and branch to
                                \ after DO if loop count not complete
                add     0 [ebp], ebx
                pop     ebx
                jo      short @@1
                add     esi, 0 [esi]
                jmp     short @@2
@@1:            add     esi, # 4
                add     ebp, # 12
@@2:            next    c;

NCODE UNLOOP    ( -- )          \ discard LOOP parameters from return stack
                add     ebp, # 12
                next    c;

NCODE (DO)      ( n1 n2 -- )    \ "runtime" setup loop using n1,n2
                pop     ecx
LABEL pdo1      xchg    esp, ebp
                mov     eax, 0 [esi]
                add     eax, esi
                push    eax
                add     esi, # 4
                add     ecx, # 0x80000000
                push    ecx
                sub     ebx, ecx
                push    ebx
                xchg    esp, ebp
                pop     ebx
                next    c;

NCODE (?DO)     ( n1 n2 -- )    \ "runtime" setup loop using n1,n2, if n1=n2
                                \ then discard n1,n2 and branch to after DO
                pop     ecx
                cmp     ecx, ebx
                jne     pdo1
                pop     ebx
                add     esi, 0 [esi]
                next    c;

CODE BOUNDS     ( adr len -- lim first ) \ calculate loop bounds from adr,len
                add     ebx, 0 [esp]
                xchg    ebx, 0 [esp]
                next    c;

CODE I          ( -- n1 )       \ push n1, the value of the current loop index
                push    ebx
                mov     ebx, 0 CELLS [ebp]
                add     ebx, 1 CELLS [ebp]
                next    c;

CODE J          ( -- n1 )       \ push n1, the value of the outer loop index
                push    ebx
                mov     ebx, 3 CELLS [ebp]
                add     ebx, 4 CELLS [ebp]
                next    c;

CODE K          ( -- n1 )       \ push n1, value of the secong outer loop index
                push    ebx
                mov     ebx, 6 CELLS [ebp]
                add     ebx, 7 CELLS [ebp]
                next    c;

NCODE LEAVE     ( -- )          \ exit the current DO LOOP immediately
                mov     esi, 2 CELLS [ebp]
                add     ebp, # 3 CELLS
                next    c;

NCODE ?LEAVE    ( f1 -- )       \ exit the current DO LOOP if f1=TRUE
                test    ebx, ebx
                pop     ebx
                je      @@1
                mov     esi, 2 CELLS [ebp]
                add     ebp, # 3 CELLS
@@1:            next    c;


\ -------------------- Stack Operators --------------------

CODE DROP       ( n -- )        \ discard top entry on data stack
                pop     ebx
                next    c;

CODE DUP        ( n -- n n )    \ duplicate top entry on data stack
                push    ebx
                next    c;

CODE SWAP       ( n1 n2 -- n2 n1 ) \ exchange first and second items on data stack
                mov     eax, 0 [esp]
                mov     0 [esp], ebx
                mov     ebx, eax
                next    c;

CODE OVER       ( n1 n2 -- n1 n2 n1 ) \ copy second item to top of data stack
                push    ebx
                mov     ebx, 4 [esp]
                next    c;

CODE ROT        ( n1 n2 n3 -- n2 n3 n1 ) \ rotate third item to top of data stack
\ fast on a Pentium, slow on a 386
                mov     eax, 4 [esp]
                mov     ecx, 0 [esp]
                mov     0 [esp], ebx
                mov     4 [esp], ecx
                mov     ebx, eax
                
\ slow on a Pentium, fast on a 386
\                pop     eax
\                pop     ecx
\                push    eax
\                push    ebx
\                mov     ebx, ecx
                next    c;

CODE -ROT       ( n1 n2 n3 -- n3 n1 n2 ) \ rotate top of data stack to third item
                mov     eax, 0 [esp]
                mov     ecx, 4 [esp]
                mov     4 [esp], ebx
                mov     0 [esp], ecx
                mov     ebx, eax
                next    c;

CODE ?DUP       ( n -- n [n] )  \ duplicate top of data stack if non-zero
                test    ebx, ebx
                je      short @@1
                push    ebx
@@1:            next    c;

CODE NIP        ( n1 n2 -- n2 ) \ discard second item on data stack
                pop     eax
                next    c;

CODE TUCK       ( n1 n2 -- n2 n1 n2 ) \ copy top data stack to under second item
                pop     eax
                push    ebx
                push    eax
                next    c;

CODE PICK       ( ...  k -- ... n[k] )
                mov     ebx, 0 [esp] [ebx*4]      \ just like that!
                next    c;

CODE SP@        ( -- addr ) \ get addr, the pointer to the top item on data stack
                push    ebx
                mov     ebx, esp
                sub     ebx, edi
                next    c;

CODE SP!        ( addr -- )     \ set the data stack to point to addr
                add     ebx, edi
                mov     esp, ebx
                pop     ebx
                next    c;

\ -------------------- Return Stack --------------------

CODE RP@        ( -- a1 )       \ get a1 the address of the return stack
                push    ebx
                mov     ebx, ebp
                sub     ebx, edi
                next    c;

CODE RP!        ( a1 -- )       \ set the address of the return stack
                add     ebx, edi
                mov     ebp, ebx
                pop     ebx
                next    c;

CODE >R         ( n1 -- )       \ push n1 onto the return stack
                sub     ebp, # 4
                mov     0 [ebp], ebx
                pop     ebx
                next    c;

CODE R>         ( -- n1 )       \ pop n1 off the return stack
                push    ebx
                mov     ebx, 0 [ebp]
                add     ebp, # 4
                next    c;

CODE R@         ( -- n1 )       \ get a copy of the top of the return stack
                push    ebx
                mov     ebx, 0 [ebp]
                next    c;

CODE DUP>R      ( n1 -- n1 )    \ push a copy of n1 onto the return stack
                sub     ebp, # 4
                mov     0 [ebp], ebx
                next    c;

CODE R>DROP     ( -- )          \ discard one item off of the return stack
                add     ebp, # 4
                next    c;

CODE 2>R        ( n1 n2 -- )    \ push a two items onto the returnstack
                pop     eax
                sub     ebp, # 8
                mov     1 CELLS [ebp], eax
                mov     0 CELLS [ebp], ebx
                pop     ebx
                next    c;

CODE 2R>        ( -- n1 n2 )    \ pop a two items off the return stack
                push    ebx
                mov     ebx, 0 CELLS [ebp]
                mov     eax, 1 CELLS [ebp]
                add     ebp, # 8
                push    eax
                next    c;

CODE 2R@        ( -- n1 n2 )    \ get a copy of the top two items on the return stack
                push    ebx
                mov     ebx, 0 CELLS [ebp]
                mov     eax, 1 CELLS [ebp]
                push    eax
                next    c;

\ -------------------- Memory Operators --------------------

CODE @          ( a1 -- n1 )    \ get the cell n1 from address a1
                mov     ebx, 0 [ebx] [edi]
                next    c;

CODE !          ( n1 a1 -- )    \ store cell n1 into address a1
                pop     eax
                mov     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE +!         ( n1 a1 -- )    \ add cell n1 to the contents of address a1
                pop     eax
                add     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE C@         ( a1 -- c1 )    \ fetch the character c1 from address a1
                mov     bl, 0 [ebx] [edi]
                and     ebx, # 0xFF
                next    c;

CODE C!         ( c1 a1 -- )    \ store character c1 into address a1
                pop     eax
                mov     0 [ebx] [edi], al
                pop     ebx
                next    c;

CODE C+!        ( c1 a1 -- )    \ add character c1 to the contents of address a1
                pop     eax
                add     0 [ebx] [edi], al
                pop     ebx
                next    c;

CODE W@         ( a1 -- w1 )    \ fetch the word (16bit) w1 from address a1
                movzx   ebx, word ptr 0 [ebx] [edi]
                next    c;

CODE SW@        ( a1 -- w1 )    \ fetch the sign extended word (16bit) w1 from address a1
                movsx   ebx, word ptr 0 [ebx] [edi]
                next    c;

CODE W!         ( w1 a1 -- )    \ store word (16bit) w1 into address a1
                pop     eax
                16:
                mov     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE W+!        ( w1 a1 -- )    \ add word (16bit) w1 to the contents of address a1
                pop     eax
                16:
                add     0 [ebx] [edi], eax
                pop     ebx
                next    c;


\ -------------------- Cell Operators --------------------

 4 CONSTANT CELL                        \ a cell is four bytes
-4 CONSTANT -CELL                       \ a negative cell is minus four bytes

CODE CELLS      ( n1 -- n1*cell )       \ multiply n1 by the cell size
                lea     ebx, 0 [ebx*4]
                next    c;

CODE CELLS+     ( a1 n1 -- a1+n1*cell ) \ multiply n1 by the cell size and add
                                        \ the result to address a1
                lea     ebx, 0 [ebx*4]
                pop     eax
                add     ebx, eax
                next    c;

CODE CELL+      ( a1 -- a1+cell )       \ add a cell to a1
                add     ebx, # 4
                next    c;

CODE CELL-      ( a1 -- a1-cell )       \ subtract a cell from a1
                sub     ebx, # 4
                next    c;

CODE +CELLS     ( n1 a1 -- n1*cell+a1 ) \ multiply n1 by the cell size and add
                                        \ the result to address a1
                pop     eax
                lea     eax, 0 [eax*4]
                add     ebx, eax
                next    c;

CODE -CELLS     ( n1 a1 -- a1-n1*cell ) \ multiply n1 by the cell size and
                                        \ subtract the result from address a1
                pop     eax
                lea     eax, 0 [eax*4]
                sub     ebx, eax
                next    c;

\ -------------------- Char Operators --------------------

CODE CHARS      ( n1 -- n1*char )       \ multiply n1 by the character size (1)
                next    c;

CODE CHAR+      ( a1 -- a1+char )       \ add the characters size in bytes to a1
                add     ebx, # 1
                next    c;

\ -------------------- Address Conversion --------------------

CODE REL>ABS    ( rel -- abs )  \ convert from a relative address as used by
                                \ Forth, to an absolute address as used by Windows
                add     ebx, edi
                next    c;

CODE ABS>REL    ( abs -- rel )  \ convert from an absolute address as used by
                                \ Windows to a relative address as used by Forth
                sub     ebx, edi
                next    c;


\ -------------------- Block Memory Operators --------------------

CODE MOVE       ( source dest len -- )  \ move len bytes from source address to
                                        \ dest address
                mov     ecx, ebx
                mov     eax, esi
                mov     ebx, edi        \ ebx = base
                pop     edi
                pop     esi
                push    edx             \ save UP
                add     esi, ebx        \ absolute address
                add     edi, ebx        \ ditto
                mov     edx, edi        \ check for overlap
                sub     edx, esi
                cmp     edx, ecx
                jb      short @@1
                mov     edx, ecx        \ CMOVE case (optimized)
                shr     ecx, # 2
                rep     movsd
                mov     ecx, edx
                and     ecx, # 3
                rep     movsb
                jmp     short @@2
@@1:            add     edi, ecx        \ CMOVE> case (dull)
                dec     edi
                add     esi, ecx
                dec     esi
                std
                rep     movsb
                cld
@@2:            mov     esi, eax
                mov     edi, ebx
                pop     edx             \ restore UP
                pop     ebx
                next    c;

CODE FILL       ( addr len char -- )    \ fill addr with char for len bytes
\                mov     al, bl          \ al = char

                mov     bh, bl          \ bh & bl = char
                shl     ebx, # 16
                mov     eax, ebx
                shr     eax, # 16
                or      eax, ebx

                mov     ebx, edi        \ ebx = base
                pop     ecx             \ ecx = len
                pop     edi             \ edi = addr
                add     edi, ebx        \ absolute address

\                shl     eax, # 8        \ fill eax with 4 characters
\                mov     al, ah
\                shl     eax, # 8
\                mov     al, ah
\                shl     eax, # 8
\                mov     al, ah

                push    ecx             \ optimize
                shr     ecx, # 2
                rep     stosd
                pop     ecx
                and     ecx, # 3
                rep     stosb
                mov     edi, ebx        \ restore
                pop     ebx
                next    c;

: ERASE         ( addr len -- )         \ fill addr for len bytes with zero
                0 FILL ;

CODE CMOVE      (  from to count -- )   \ move bytes "from" -> "to" of "count"
                mov     ecx, ebx        \ start at first byte of "from"
                mov     eax, esi
                mov     ebx, edi
                pop     edi
                pop     esi
                push    edx             \ save UP
                add     edi, ebx
                add     esi, ebx
                repnz   movsb
                mov     esi, eax
                mov     edi, ebx
                pop     edx             \ restore UP
                pop     ebx
                next    c;

CODE CMOVE>     ( from to count -- )    \ move bytes "from" -> "to" of "count"
                mov     ecx, ebx        \ start at last byte of "from"
                mov     eax, esi
                mov     ebx, edi
                dec     ecx
                pop     edi
                pop     esi
                push    edx             \ save UP
                add     edi, ecx
                add     edi, ebx
                add     esi, ecx
                add     esi, ebx
                inc     ecx
                std
                repnz   movsb
                cld
                mov     esi, eax
                mov     edi, ebx
                pop     edx             \ restore UP
                pop     ebx
                next    c;

\ -------------------- Logical Operators --------------------

CODE AND        ( n1 n2 -- n3 ) \ perform bitwise AND of n1,n2, return result n3
                pop     eax
                and     ebx, eax
                next    c;

CODE OR         ( n1 n2 -- n3 ) \ perform bitwise OR of n1,n2, return result n3
                pop     eax
                or      ebx, eax
                next    c;

CODE XOR        ( n1 n2 -- n3 ) \ perform bitwise XOR of n1,n2, return result n3
                pop     eax
                xor     ebx, eax
                next    c;

CODE INVERT     ( n1 -- n2 )    \ perform a bitwise -1 XOR on n1, return result n2
                not     ebx
                next    c;

CODE LSHIFT     ( u1 n -- u2 )  \ shift u1 left by n bits (multiply)
                mov     ecx, ebx
                pop     ebx
                shl     ebx, cl
                next    c;

CODE RSHIFT     ( u1 n -- u2 )  \ shift u1 right by n bits (divide)
                mov     ecx, ebx
                pop     ebx
                shr     ebx, cl
                next    c;

-1 CONSTANT TRUE   ( -- true )  \ put TRUE flag on the stack
 0 CONSTANT FALSE  ( -- false ) \ put FALSE flag on the stack

CODE INCR       ( addr -- )     \ increment the contents of addr
                mov     eax, # 1
                add     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE DECR       ( addr -- )     \ decrement the contents of addr
                mov     eax, # -1
                add     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE CINCR      ( addr -- )     \ increment the BYTE contents of addr
                mov     eax, # 1
                add     0 [ebx] [edi], al
                pop     ebx
                next    c;

CODE CDECR      ( addr -- )     \ decrement the BYTE contents of addr
                mov     eax, # -1
                add     0 [ebx] [edi], al
                pop     ebx
                next    c;

CODE ON         ( addr -- )     \ set the contents of addr to ON (-1)
                mov     eax, # -1
                mov     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE OFF        ( addr -- )     \ set the contests of addr of OFF (0)
                sub     eax, eax
                mov     0 [ebx] [edi], eax
                pop     ebx
                next    c;

CODE TOGGLE     ( addr byte -- ) \ XOR the byte contents of "addr" with "byte"
                pop     eax
                xor     0 [eax] [edi], bl
                pop     ebx
                next    c;

\ -------------------- Arithmetic Operators --------------------

CODE +          ( n1 n2 -- n3 ) \ add n1 to n2, return sum n3
                pop     eax
                add     ebx, eax
                next    c;

CODE -          ( n1 n2 -- n3 ) \ subtract n2 from n1, return difference n3
                pop     eax
                sub     eax, ebx
                mov     ebx, eax
                next    c;

CODE UNDER+     ( a x b -- a+b x ) \ add top of stack to third stack item
                add     4 [esp], ebx
                pop     ebx
                next    c;

CODE NEGATE     ( n1 -- n2 ) \ negate n1, returning 2's complement n2
                neg     ebx
                next    c;

CODE ABS        ( n -- |n| ) \ return the absolute value of n1 as n2
                test    ebx, ebx
                jns     @@1
                neg     ebx
@@1:            next    c;

0 CONSTANT 0
1 CONSTANT 1

CODE 2*         ( n1 -- n2 ) \ multiply n1 by two
                lea     ebx, 0 [ebx*2]
                next    c;

CODE 2/         ( n1 -- n2 ) \ signed divide n1 by two
                sar     ebx, # 1
                next    c;

CODE U2/        ( n1 -- n2 ) \ unsigned divide n1 by two
                shr     ebx, # 1
                next    c;

CODE 1+         ( n1 -- n2 ) \ add one to n1
                inc     ebx
                next    c;

CODE 1-         ( n1 -- n2 ) \ subtract one from n1
                dec     ebx
                next    c;

CODE D2*        ( d1 -- d2 ) \ multiply the double number d1 by two
                pop     eax
                shl     eax, # 1
                rcl     ebx, # 1
                push    eax
                next    c;

CODE D2/        ( d1 -- d2 ) \ divide the double number d1 by two
                pop     eax
                sar     ebx, # 1
                rcr     eax, # 1
                push    eax
                next    c;

\ -------------------- Unsigned Multiply & Divide --------------------

CODE UM*        ( u1 u2 -- ud1 ) \ multiply unsigned u1 by unsigned u2
                mov     ecx, edx        \ save UP
                pop     eax
                mul     ebx
                push    eax
                mov     ebx, edx
                mov     edx, ecx        \ restore UP
                next    c;

CODE UM/MOD     ( ud1 u1 -- rem quot ) \ divide unsigned double ud1 by the
                                       \ unsigned number u1
                mov     ecx, edx        \ save UP
                pop     edx
                pop     eax
                div     ebx
                push    edx
                mov     ebx, eax
                mov     edx, ecx        \ restore UP
                next    c;

CODE WORD-SPLIT ( u1 -- low high ) \ split the unsigned 32bit u1 into its high
                                   \ and low 16bit quantities.
                sub     eax, eax
                16:
                mov     eax, ebx
                push    eax
                shr     ebx, # 16
                next    c;

CODE WORD-JOIN  ( low high -- n1 ) \ join the high and low 16bit quantities
                                   \ into a single 32bit n1
                shl     ebx, # 16
                pop     eax
                16:
                mov     ebx, eax
                next    c;

\ -------------------- Comparison Operators --------------------

CODE 0=         ( n1 -- f1 )    \ return true if n1 equals zero
                sub     ebx, # 1
                sbb     ebx, ebx
                next    c;

CODE 0<>        ( n1 -- f1 )    \ return true if n1 is not equal to zero
                sub     eax, eax
                sub     eax, ebx
                neg     eax
                sbb     ebx, ebx
                next    c;

CODE 0<         ( n1 -- f1 )    \ return true if n1 is less than zero
                sar     ebx, # 31
                next    c;

CODE 0>         ( n1 -- f1 )    \ return true if n1 is greater than zero
                cmp     ebx, # 0
                jle     short @@1
                mov     ebx, # -1
                jmp     short @@2
@@1:            sub     ebx, ebx
@@2:            next    c;

CODE =          ( n1 n2 -- f1 ) \ return true if n1 is equal to n2
                pop     eax
                sub     ebx, eax
                sub     ebx, # 1
                sbb     ebx, ebx
                next    c;

CODE <>         ( n1 n2 -- f1 ) \ return true if n1 is not equal to n2
                pop     eax
                sub     eax, ebx
                neg     eax
                sbb     ebx, ebx
                next    c;

CODE <          ( n1 n2 -- f1 ) \ return true if n1 is less than n2
                pop     eax
                cmp     eax, ebx
                jge     short @@1
                mov     ebx, # -1
                jmp     short @@2
@@1:            sub     ebx, ebx
@@2:            next    c;

CODE >          ( n1 n2 -- f1 ) \ return true if n1 is greater than n2
                pop     eax
                cmp     eax, ebx
                jle     short @@1
                mov     ebx, # -1
                jmp     short @@2
@@1:            sub     ebx, ebx
@@2:            next    c;

CODE U<         ( u1 u2 -- f1 ) \ return true if unsigned u1 is less than
                                \ unsigned u2
                pop     eax
                cmp     eax, ebx
                sbb     ebx, ebx
                next    c;

CODE U>         ( u1 u2 -- f1 ) \ return true if unsigned u1 is greater than
                                \ unsigned n2
                pop     eax
                cmp     ebx, eax
                sbb     ebx, ebx
                next    c;

CODE DU<        ( ud1 ud2 -- f1 ) \ return true if unsigned double ud1 is
                                  \ less than undigned double ud2
                pop     eax
                pop     ecx
                xchg    edx, 0 [esp]    \ save UP
                sub     edx, eax
                sbb     ecx, ebx

                sbb     ebx, ebx

\                sbb     edx, edx
\                mov     ebx, edx

                pop     edx             \ restore UP
                next c;

CODE UMIN       ( u1 u2 -- n3 ) \ return the lesser of unsigned u1 and
                                \ unsigned u2
                pop     eax
                cmp     ebx, eax
                jb      short @@1
                mov     ebx, eax
@@1:            next    c;

CODE MIN        ( n1 n2 -- n3 ) \ return the lesser of n1 and n2
                pop     eax
                cmp     ebx, eax
                jl      short @@1
                mov     ebx, eax
@@1:            next    c;

CODE UMAX       ( u1 u2 -- n3 ) \ return the greater of unsigned u1 and
                                \ unsigned u2
                pop     eax
                cmp     ebx, eax
                ja      short @@1
                mov     ebx, eax
@@1:            next    c;

CODE MAX        ( n1 n2 -- n3 ) \ return the greater of n1 and n2
                pop     eax
                cmp     ebx, eax
                jg      short @@1
                mov     ebx, eax
@@1:            next    c;

CODE 0MAX       ( n1 -- n2 ) \ return n2 the greater of n1 and zero
                cmp     ebx, # 0
                jg      short @@1
                mov     ebx, # 0
@@1:            next    c;

CODE BETWEEN    ( n1 low high -- f1 ) \ f1=true if ((n1 >= low) & (n1 <= high))
                pop     ecx
                pop     eax
                cmp     eax, ecx
                jl      short @@1
                cmp     eax, ebx
                jg      short @@1
                mov     ebx, # -1
                jmp     short @@2
@@1:            sub     ebx, ebx
@@2:            next    c;

CODE WITHIN     ( n1 low high -- f1 ) \ f1=true if ((n1 >= low) & (n1 < high))
                pop     eax
                pop     ecx
                sub     ebx, eax
                sub     ecx, eax
                sub     ecx, ebx
                sbb     ebx, ebx
                next    c;

\ -------------------- Double memory Operators --------------------

CODE 2@         ( a1 -- d1 ) \ fetch the double number n1 from address a1
                mov     eax, 4 [ebx] [edi]
                mov     ebx, 0 [ebx] [edi]
                push    eax
                next    c;

CODE 2!         ( d1 a1 -- ) \ store the double number d1 into address a1
                pop     eax
                mov     0 [ebx] [edi], eax
                pop     eax
                mov     4 [ebx] [edi], eax
                pop     ebx
                next    c;

\ -------------------- Double Stack Operators --------------------

CODE 2DROP      ( n1 n2 -- ) \ discard two single items from the data stack
                pop     ebx
                pop     ebx
                next    c;

CODE 2DUP       ( n1 n2 -- n1 n2 n1 n2 ) \ duplicate the top two single items
                                         \ on the data stack
                mov     eax, 0 [esp]
                push    ebx
                push    eax
                next    c;

CODE 2SWAP      ( n1 n2 n3 n4 -- n3 n4 n1 n2 )
                pop     eax
                xchg    ebx, 0 [esp]
                xchg    eax, 4 [esp]
                push    eax
                next    c;

CODE 2OVER      ( n1 n2 n3 n4 -- n1 n2 n3 n4 n1 n2 )
                mov     eax, 8 [esp]
                push    ebx
                push    eax
                mov     ebx, 12 [esp]
                next    c;

CODE 2ROT       ( n1 n2 n3 n4 n5 n6 -- n3 n4 n5 n6 n1 n2)
                pop     eax
                xchg    ebx, 0 [esp]
                xchg    eax, 4 [esp]
                xchg    ebx, 8 [esp]
                xchg    eax, 12 [esp]
                push    eax
                next    c;

CODE 3DROP      ( n1 n2 n3 -- ) \ discard three items from the data stack
                add     esp, # 8
                pop     ebx
                next    c;

CODE 4DROP      ( n1 n2 n3 n4 -- ) \ discard four items from the data stack
                add     esp, # 12
                pop     ebx
                next    c;

: 3DUP          ( n1 n2 n3 -- n1 n2 n3 n1 n2 n3 )
\ Duplicate top 3 single numbers on the stack.
                >R 2DUP R@ -ROT R> ;

: 4DUP          ( a b c d -- a b c d a b c d )
\ Duplicate top 4 single numbers (or two double numbers) on the stack.
                2OVER 2OVER   ;

\ -------------------- Double Arithmetic Operators --------------------

CODE D+         ( d1 d2 -- d3 ) \ perform a double add (64bit)
                pop     eax
                pop     ecx
                add     0 [esp], eax
                adc     ebx, ecx
                next    c;

CODE D-         ( d1 d2 -- d3 ) \ perform a double subtract (64bit)
                pop     eax
                pop     ecx
                sub     0 [esp], eax
                sbb     ecx, ebx
                mov     ebx, ecx
                next    c;

CODE DNEGATE    ( d1 -- d2 )    \ negate d1, returning 2's complement d2
                pop     eax
                neg     ebx
                neg     eax
                sbb     ebx, # 0
                push    eax
                next    c;

CODE DABS       ( d1 -- d2 )    \ return the absolute value of d1 as d2
                test    ebx, ebx
                jns     @@1
                pop     eax
                neg     ebx
                neg     eax
                sbb     ebx, # 0
                push    eax
@@1:            next    c;

CODE S>D        ( n1 -- d1 ) \ convert single signed single n1 to a signed
                             \ double d1
                push    ebx
                shl     ebx, 1          \ put sign bit into carry
                sbb     ebx, ebx
                next    c;

: D=            ( d1 d2 -- f1 ) \ f1=true if double d1 is equal to double d2
                D- OR 0= ;

CODE D0<        ( d1 -- f1 )
\ Signed compare d1 double number with zero.  If d1 < 0, return TRUE.
                sar     ebx, # 31
                pop     ecx
                next    c;

CODE D<         ( d1 d2 -- f )
\ Signed compare two double numbers.  If d1 < d2, return TRUE.
                pop     eax
                pop     ecx
                cmp     0 [esp], eax
                pop     eax
                sbb     ecx, ebx
                mov     ebx, # 0
                jge     short @@1
                dec     ebx
@@1:            next    c;


: D>            ( d1 d2 -- f )
\ Signed compare two double numbers.  If d1 > d2 , return TRUE.
                2SWAP D<   ;

: D<>           ( d1 d2 -- d )
\ Signed compare two double numbers.  If d1 not equal d2 , return TRUE.
                D= 0= ;

: DMIN          ( d1 d2 -- d3 )
\  Replace the top two double numbers with the smaller of the two (signed).
                4DUP D> IF  2SWAP  THEN 2DROP ;

: DMAX          ( d1 d2 -- d3 )
\  Replace the top two double numbers with the larger of the two (signed).
                4DUP D< IF  2SWAP  THEN  2DROP ;        \ 05/25/90 tjz

\ -------------------- Signed Multiply & Divide --------------------

CODE M*         ( n1 n2 -- d1 ) \ multiply n1 by n2, return double result d1
                mov     ecx, edx        \ save UP
                pop     eax
                imul    ebx
                push    eax
                mov     ebx, edx
                mov     edx, ecx        \ restore UP
                next    c;

CODE *          ( n1 n2 -- n3 ) \ multiply n1 by n2, return single result n3
                mov     ecx, edx        \ save UP
                pop     eax
                mul     ebx
                mov     ebx, eax
                mov     edx, ecx        \ restore UP
                next    c;

CODE SM/REM     ( d n -- rem quot )
                mov     ecx, edx        \ save UP
                pop     edx
                pop     eax
                idiv    ebx
                push    edx
                mov     ebx, eax
                mov     edx, ecx        \ restore UP
                next    c;

CODE FM/MOD     ( d n -- rem quot )
                pop     ecx             \ high numerator
                mov     eax, ecx        \ copy for testing
                xor     eax, ebx        \ test against denominator
                jns     short @@1       \ if signs differ jump

                xchg    ecx, edx        \ save UP
                pop eax
                idiv    ebx
                test    edx, edx        \ set zero flag
                je      short @@2
                add     edx, ebx        \ add divisor to remainder
                dec     eax             \ decrement quotient
                jmp     short @@2

@@1:            xchg    ecx, edx        \ preserve DX in CX, DX=high num
                pop     eax             \ EAX=low numerator
                idiv    ebx             \ perform the division

@@2:            push    edx             \ push remainder
                mov     ebx, eax        \ quotient to EBX
                mov     edx, ecx        \ restore UP
                next    c;

CODE /MOD       ( n1 n2 -- rem quot )
                pop     ecx
                mov     eax, ecx
                xor     eax, ebx
                jns     short @@1

                mov     eax, ecx        \ low order part to eax
                mov     ecx, edx        \ save UP
                cdq
                idiv    ebx
                test    edx, edx        \ set zero flag
                je      short @@2

                add     edx, ebx        \ add divisor to remainder
                dec     eax             \ decrement quotient
                jmp     short @@2

@@1:            mov     eax, ecx        \ low order part to eax
                mov     ecx, edx
                cdq
                idiv    ebx
@@2:            push    edx
                mov     ebx, eax
                mov     edx, ecx        \ restore UP
                next    c;

: /             ( n1 n2 -- n3 )
                /MOD NIP ;

: MOD           ( n1 n2 -- n3 )
                /MOD DROP ;

CODE */MOD      ( n1*n2/n3 -- remainder quotient)
                pop     ecx
                pop     eax
                push    edx             \ save UP
                imul    ecx
                mov     ecx, edx
                xor     ecx, ebx
                jns     short @@1

                idiv    ebx
                test    edx, edx        \ set zero flag
                je      short @@2
                add     edx, ebx        \ add divisor to remainder
                dec     eax             \ decrement quotient
                jmp     short @@2

@@1:            idiv    ebx
@@2:            mov     ebx, eax
                pop     ecx
                push    edx
                mov     edx, ecx        \ restore UP
                next    c;

: */            ( n1*n2/n3 -- quotient )
                */MOD NIP ;

\ -------------------- User Variables --------------------

4096 CONSTANT USERSIZE

VARIABLE NEXT-USER              \ offset of next user variable

CREATE CONUSER  ( -- tid )      \ the console user area
                USERSIZE ALLOT

CODE UP@        ( -- addr )     \ get the pointer to the user area
                push    ebx
                mov     ebx, edx
                sub     ebx, edi
                next    c;

CODE UP!        ( addr -- )     \ set the pointer to the user area
                lea     edx, 0 [ebx] [edi]      \ make absolute user base
                pop     ebx
                next    c;

-3 CELLS ( init offset so FOLLOWER is 0 ) ( order IS important )
  DUP USER RP0        CELL+ ( initial return stack pointer )
  DUP USER SP0        CELL+ ( initial data stack pointer )
  DUP USER STATUS     CELL+ ( PASS or WAKE )
  DUP USER FOLLOWER   CELL+ ( address of next task's STATUS )
  DUP USER TOS        CELL+ ( save top of data stack )
  DUP ( *1 )                ( absolute min user area )
  DUP USER HANDLER    CELL+ ( throw frame )
  DUP USER LP         CELL+ ( local variable pointer )
  DUP USER OP         CELL+ ( object pointer )
  DUP USER BASE       CELL+ ( numeric radix )
  DUP USER HLD        CELL+ ( numeric output pointer )
         80 CHARS ALIGNED + ( numeric output formatting buffer )
  DUP USER PAD        CELL+ ( extra )
  MAXSTRING CHARS ALIGNED + ( user string buffer )
  DUP USER MSG        CELL+ ( abort message pointer )
  DUP NEXT-USER !-T                 ( save top user variable offset in NEXT-USER )
( *2 )    OVER - CONSTANT USEREXTRA ( add to USERMIN if you are going to do I/O )
( *1 ) 3 CELLS + CONSTANT USERMIN   ( absolute min user area )

\ -------------------- System Variables --------------------

CREATE POCKET           MAXSTRING ALLOT
CREATE CUR-FILE         MAXSTRING ALLOT
CREATE TEMP$            MAXSTRING ALLOT
CREATE FIND-BUFFER      MAXSTRING ALLOT

VARIABLE DP                             \ application dictionary pointer
VARIABLE SDP                            \ system dictionary pointer

NUMVOCS CONSTANT #VOCS

VARIABLE CONTEXT  #VOCS CELLS ALLOT     \ make context array of #VOCS+1 cells
VARIABLE CURRENT
VARIABLE LAST                   \ NFA of alst header created
VARIABLE LAST-LINK              \ address of last link for last header created
VARIABLE VOC-LINK               \ linked list of vocabularies

0 VALUE SOURCE-ID
0 VALUE SOURCE-POSITION

CREATE TIB                      \ terminal input buffer
        MAXSTRING ALLOT

CREATE (SOURCE)                 \ input stream pointer
        0 ,                     \ length of input stream
        TIB ,                   \ address of input stream

(SOURCE) CONSTANT #TIB          \ address of terminal input buffer length

(SOURCE)     EQU S_LEN          \ source address & length for the assembler
(SOURCE) 4 + EQU S_ADR

VARIABLE >IN
VARIABLE STATE
VARIABLE WARNING TRUE WARNING !-T
VARIABLE CAPS    TRUE CAPS    !-T

CREATE .SMAX   8 ,              \ max number of stack entries to show

VARIABLE LOADLINE
VARIABLE ?LOADING

\ -------------------- Strings --------------------

32 CONSTANT BL

: BLANK         ( addr len -- )     \ fill addr for len with spaces (blanks)
                BL FILL ;

CODE COUNT      ( str -- addr len ) \ byte counted strings
                sub     eax, eax
                mov     al, 0 [ebx] [edi]
                inc     ebx
                push    ebx
                mov     ebx, eax
                next    c;

CODE WCOUNT     ( str -- addr len )  \ word counted strings
                sub     eax, eax
                16:
                mov     eax, 0 [ebx] [edi]
                inc     ebx
                inc     ebx
                push    ebx
                mov     ebx, eax
                next    c;

CODE LCOUNT     ( str -- addr len )  \ long counted strings
                sub     eax, eax
                mov     eax, 0 [ebx] [edi]
                add     ebx, # 4
                push    ebx
                mov     ebx, eax
                next    c;

: "CLIP"        ( a1 n1 -- a1 n1 )   \ clip a string to between 0 and MAXCOUNTED
                MAXCOUNTED MIN 0 MAX ;

: PLACE         ( addr len dest -- )
                SWAP "CLIP" SWAP
                2DUP 2>R
                CHAR+ SWAP MOVE
                2R> C! ;

: +PLACE        ( addr len dest -- ) \ append string addr,len to counted
                                     \ string dest
                >R "CLIP" MAXCOUNTED R@ C@ - MIN R>
                                        \ clip total to MAXCOUNTED string
                2DUP 2>R
                COUNT CHARS + SWAP MOVE
                2R> C+! ;

CODE -TRAILCHARS ( addr n1 c1 -- addr n2 ) \ remove trailing c1's from addr,n1
                mov     eax, ebx
                pop     ecx
                jecxz   short @@2
                mov     ebx, 0 [esp]
                add     ebx, ecx
@@1:            dec     ebx
                cmp     0 [ebx] [edi], al
                jne     short @@2
                dec     ecx
                jnz     @@1
@@2:            mov     ebx, ecx
                next    c;

: -TRAILING     ( addr n1 -- addr n2 ) \ remove trailing blanks from addr,n1
                BL -TRAILCHARS ;

: -NULLS        ( addr n1 -- addr n2 ) \ remove trailing nulls from addr,n1
                0 -TRAILCHARS ;

\ remove n1 characters from the beginning of string addr1,len1
\ if n1 greater than len1, then returned len2 will be zero.

\ ** November 27th, 2000 - 12:00 tjz
\ ** the following statment is no longer true **
\ if n1 lessthan zero, then returned len2 will be zero.

CODE /STRING    ( addr1 len1 n1 -- addr2 len2 )
                pop     eax
// November 27th, 2000 - 11:58 tjz
// Added the following two lines to allow a negative argument to be passed
// to /STRING, such that the string will be expanded.  Dangerous, but legal.
                cmp     ebx, # 0
                jle     short @@1

                cmp     ebx, eax
                jbe     short @@1
                mov     ebx, eax
@@1:            add     0 [esp], ebx
                sub     eax, ebx
                mov     ebx, eax
                next    c;

\ -------------------- External C Calls --------------------

0       equ exit_x
1       equ initconsole_x
2       equ key_x
3       equ accept_x
4       equ emit_x
5       equ type_x
6       equ cr_x
7       equ cls_x
8       equ keyq_x
9       equ qcr_x
10      equ malloc_x
11      equ free_x
12      equ open_file_x
13      equ close_file_x
14      equ read_line_x
15      equ save_x
16      equ load_library_x
17      equ freelibrary_x
18      equ get_proc_x
19      equ gotoxy_x
20      equ getxy_x
\ 21      equ fconvert_x        \ August 21st, 1996 - 17:32 obsolete don't use
\ 22      equ fformat_x         \ August 21st, 1996 - 17:32 obsolete don't use
23      equ getcolrow_x
24      equ getstartup_x
25      equ tone_x
26      equ beepstore_x
27      equ setfgbg_x
28      equ pushkey_x
29      equ thescreen_x         \ 06/22/94 10:01 tjz added
30      equ charwh_x            \ 06/22/94 10:43 tjz added
31      equ shiftmask_x         \ 06/22/94 11:49 tjz added
32      equ resize_x            \ 07/06/94 15:56 tjz added
33      equ reposition_x        \ 07/06/94 16:36 tjz added
34      equ pointershape_x      \ 07/08/94 11:34 tjz added
35      equ cursorshape_x       \ 07/12/94 10:47 tjz added
36      equ getcursor_x         \ 07/12/94 12:42 tjz added
37      equ havemenu_x          \ 07/26/94 17:51 tjz added
38      equ condc_x             \ 08/03/94 14:48 tjz added
39      equ printdlg_x          \ 08/15/94 17:13 tjz added
40      equ startprint_x        \ 08/15/94 17:13 tjz added
41      equ pageprint_x         \ 08/16/94 13:29 tjz added
42      equ endprint_x          \ 08/16/94 13:29 tjz added
43      equ initprint_x         \ 08/16/94 14:55 tjz added
44      equ closeprint_x        \ 08/16/94 15:01 tjz added
45      equ opendialog_x        \ 08/19/94 17:43 tjz added
46      equ savedialog_x        \ 08/23/94 09:03 tjz added
47      equ realloc_x           \ 08/30/94 09:42 tjz added
48      equ hideconsole_x       \ 09/14/94 09:24 tjz added
49      equ showconsole_x       \ 09/14/94 09:25 tjz added
50      equ setfocus_x          \ 09/14/94 09:44 tjz added
51      equ wscroll_x           \ 09/15/94 15:32 tjz added
52      equ rowoffset_x         \ 09/15/94 16:21 tjz added
53      equ maxcolrow_x         \ 09/15/94 16:21 tjz added
54      equ setmaxcolrow_x      \ 09/20/94 10:28 tjz added
55      equ newdialog_x         \ 09/27/94 09:52 tjz added
56      equ normconsole_x       \ 09/27/94 09:51 tjz added
57      equ qualityprint_x      \ 10/03/94 14:55 tjz added
58      equ startpage_x         \ 12/16/94 09:12 tjz added
59      equ endpage_x           \ 12/16/94 09:13 tjz added
60      equ printcopies_x       \ 12/30/94 15:22 tjz added
61      equ printfrompages_x    \ 12/30/94 15:23 tjz added
62      equ printtopages_x      \ 12/30/94 15:23 tjz added
63      equ seek_file_x         \ 10/24/95 9:26 tjz added
64      equ make_file_x         \ 11/10/95 15:16 tjz added
65      equ read_file_x         \ 11/10/95 15:16 tjz added
66      equ sizestate_x         \ March 11th, 1996 - 10:39 tjz added
67      equ orientation_x       \ April 17th, 1996 - 17:44 yjz added
68      equ uppercase_x         \ September 12th, 1996 - 9:34 tjz added
69      equ lowercase_x         \ September 12th, 1996 - 9:35 tjz added
70      equ file_position_x     \ December 6th, 1996 - 15:13 tjz added
71      equ setcharwh_x         \ December 9th, 1996 - 8:31 tjz added
72      equ mark_x              \ May 20th, 1997 - 18:40 tjz added
73      equ getfg_x             \ March 1st, 2000 - 15:49 tjz added
74      equ getbg_x             \ March 1st, 2000 - 15:49 tjz added
75      equ opendialog2_x       \ January 31st, 2002 - 16:16 rls added
76      equ newdialog2_x        \ January 31st, 2002 - 16:17 rls added
77      equ initprint2_x        \ January 31st, 2002 - 16:17 rls added
78      equ printflags_x        \ February 4th, 2002 - 5:35 rls added
79      equ getfileindex_x      \ February 13th, 2002 - 5:49 rls added

CODE XCALL      ( [...] func# -- result )
                xchg    ebx, edx                \ save UP
                lea     edx, 0 [edx*4]
                add     edx, table [edi]
                call    0 [edx]
                mov     edx, ebx                \ restore UP
                mov     ebx, eax                \ function return value
                next    c;

: _BYE          ( -- )          \ Exit Forth
                0 exit_x XCALL ;

: INIT-CONSOLE  ( -- f1 )       \ initialize the Forth console window
                                \ f1=false if already inited
                HCON @ >R
                initconsole_x XCALL
                DUP HCON !
                0= THROW
                R> HCON @ - ;

: INIT-SCREEN   ( -- )          \ init the screen
                INIT-CONSOLE DROP
                normconsole_x XCALL DROP ;

: SIZESTATE     ( -- state )    \ state of the display
                sizestate_x XCALL ;

: EKEY          ( -- key )
                key_x XCALL ;

: _ACCEPT       ( addr len -- n )
                SWAP REL>ABS
                accept_x XCALL ;

: _EMIT         ( char -- )
                emit_x XCALL DROP ;

: _TYPE         ( addr len -- )
                SWAP REL>ABS type_x XCALL DROP ;

: _CR           ( -- )
                cr_x XCALL DROP ;

: _CLS          ( -- )
                cls_x XCALL DROP ;

: EKEY?         ( -- f )
                keyq_x XCALL ;

: _?CR          ( n -- )
                qcr_x XCALL DROP ;

: _MALLOC       ( n -- a1 )
                malloc_x XCALL ABS>REL ;

: _FREE         ( a1 -- )
                REL>ABS free_x XCALL DROP ;

: &START-NAME   ( -- n )
                getstartup_x XCALL ;

\ -------------------- File I/O --------------------

: FOPEN-FILE    ( addr len fam -- fileid ior )
                SWAP ROT REL>ABS open_file_x XCALL DUP 0= ;

: FMAKE-FILE    ( addr len -- fileid ior )
                SWAP REL>ABS make_file_x XCALL DUP 0= ;

: FCLOSE-FILE   ( fileid -- ior )
                close_file_x XCALL ;

: FREAD-LINE    ( addr len fileid -- len flag ior )
                SWAP ROT REL>ABS read_line_x XCALL
                DUP 0 MAX SWAP DUP -1 <> SWAP -2 = ;

: FREAD-FILE    ( addr len fileid -- len )
                ROT REL>ABS -ROT read_file_x XCALL ;

: FSEEK-FILE    ( offset origin fileid -- ior )
                seek_file_x XCALL ;

: FREPOSITION-FILE ( len-ud fileid -- ior )
                nip 0 ( SEEK_SET ) swap FSEEK-FILE ;

: FSAVE-FILE    ( addr len filename -- )
                SWAP ROT >R >R
                COUNT SWAP REL>ABS              \ file name & length
                R>                              \ length
                R> REL>ABS                      \ address
                save_x XCALL
                0= ABORT" save error" ;

\ this function only returns a single precision position, so high part is
\ always zero, and it can't fail, so error is always zero

: FFILE-POSITION ( fileid -- file_position_low file_position_high=0 f1 )
                file_position_x XCALL 0 over -1 = ;

: _GOTOXY       ( x y -- )
                SWAP gotoxy_x XCALL DROP ;

: _GETXY        ( -- x y )
                getxy_x XCALL WORD-SPLIT ;

: _GETCOLROW    ( -- cols rows )
                getcolrow_x XCALL WORD-SPLIT ;

: TONE          ( frequency duration-ms -- )
                SWAP tone_x XCALL DROP ;

: BEEP!         ( frequency duration-ms -- )
                SWAP beepstore_x XCALL DROP ;

: PRINT-ORIENTATION ( f1 -- hDC ) \ true = landscape
                orientation_x XCALL ;

: UPC           ( char -- char )        \ convert char to uppercase
                SP@ REL>ABS 1 SWAP uppercase_x XCALL DROP ;

: UPPER         ( addr len -- )         \ convert string addr,len to uppercase
                SWAP REL>ABS uppercase_x XCALL DROP ;

: LOWER         ( addr len -- )         \ convert string addr,len to lowercase
                SWAP REL>ABS lowercase_x XCALL DROP ;

: UPPERCASE     ( str -- str )
                DUP COUNT SWAP REL>ABS uppercase_x XCALL DROP ;

NCODE ?UPPERCASE ( str -- str )
                mov     cl, 0 [ebx] [edi]
                cmp     cl, # 3                 \ length 3
                jnz     short @@4
                mov     cl, 1 [ebx] [edi]
                cmp     cl, # CHAR '            \ first char is a tick
                jnz     short @@4
                mov     cl, 3 [ebx] [edi]
                cmp     cl, # CHAR '            \ third char is a tick
                jz      short @@3
@@4:            test    dword ptr CAPS [edi], # -1
                jz      short @@3
                mov     cl, 0 [ebx] [edi]
                and     ecx, # 0xFF
                jz      short @@3
                lea     eax, ' UPPERCASE        \ get address of UPPERCASE
                exec                            \ go execute it
@@3:            next    c;

: MARKCONSOLE   ( startline startcol endline endcol -- )
                mark_x XCALL DROP ;

\ ------------------- Deferred I/O -------------------

DEFER BYE       ' _BYE       IS BYE
DEFER KEY       ' EKEY       IS KEY
DEFER ACCEPT    ' _ACCEPT    IS ACCEPT
DEFER EMIT      ' _EMIT      IS EMIT
DEFER TYPE      ' _TYPE      IS TYPE
DEFER CR        ' _CR        IS CR
DEFER CLS       ' _CLS       IS CLS
DEFER KEY?      ' EKEY?      IS KEY?
DEFER ?CR       ' _?CR       IS ?CR
DEFER CONSOLE   ' NOOP       IS CONSOLE
DEFER GOTOXY    ' _GOTOXY    IS GOTOXY
DEFER GETXY     ' _GETXY     IS GETXY
DEFER GETCOLROW ' _GETCOLROW IS GETCOLROW
DEFER MS        ' DROP       IS MS

\ --------------- DLL Procedure Call ---------------

CREATE DOCALL                   \ runtime for a DLL call
                INIT-ASSEMBLER
                push    ebx
                mov     ebx, edx        \ save UP
                call    4 [eax] [edi]
                mov     edx, ebx        \ restore UP
                mov     ebx, eax
                next    c;

CODE CALL-PROC
                xchg    ebx, edx        \ save UP
                call    edx
                xchg    edx, ebx        \ restore UP
                mov     ebx, eax
                next    c;

CODE (CALL)     ( [parms] proc -- result )
                mov     eax, 0 [ebx] [edi]      \ procedure address
                mov     ebx, edx                \ save UP
                call    eax
                mov     edx, ebx                \ restore UP
                mov     ebx, eax                \ return result
                next    c;

\ -------------------- Number Input --------------------

CODE DIGIT      ( char base -- n flag )
                mov     eax, 0 [esp]
                sub     al, # 48
                jb      short @@1
                cmp     al, # 9
                jbe     short @@2
                sub     al, # 7
                cmp     al, # 10
                jb      short @@1
@@2:            cmp     al, bl
                jae     short @@1
                mov     0 [esp], eax
                mov     ebx, # -1
                jmp     short @@3
@@1:            sub     ebx, ebx
@@3:            next    c;


CODE >NUMBER    ( ud addr len -- ud addr len )
                xchg    esp, ebp                \ 0 CELLS [ebp] = addr
                push    esi                     \ 1 CELLS [ebp] = high
                push    edi                     \ 2 CELLS [esp] = low
                push    edx                     \ save UP
                mov     ecx, ebx                \ ecx = length
                jecxz   short @@4
                mov     esi, 0 [ebp]            \ esi = address
                add     esi, edi                \ absolute address
                mov     ebx, BASE [edx]         \ get the number base
@@1:            sub     eax, eax
                mov     al, 0 [esi]             \ get next digit
                cmp     al, # '0'
                jb      short @@3               \ if below '0' branch to done
                cmp     al, # '9'
                jbe     short @@2               \ go convert it
                and     al, # 0xDF              \ convert to uppercase
                cmp     al, # 'A'               \ if below 'A'
                jb      short @@3               \ then branch to done
                sub     al, # 7
@@2:            sub     al, # '0'
                cmp     al, bl
                jae     short @@3
                xchg    eax, 1 CELLS [ebp]      \ high word * base
                mul     ebx
                xchg    eax, 2 CELLS [ebp]      \ low word * base
                mul     ebx
                add     eax, 1 CELLS [ebp]      \ add
                adc     edx, 2 CELLS [ebp]
                mov     2 CELLS [ebp], eax      \ store result
                mov     1 CELLS [ebp], edx
                inc     esi
                dec     ecx
                jnz     @@1
@@3:            sub     esi, edi                \ relative address
                mov     0 CELLS [ebp], esi      \ address of unconvertable digit
                mov     ebx, ecx                \ length remaining
@@4:            pop     edx                     \ restore UP
                pop     edi
                pop     esi
                xchg    esp, ebp
                next    c;

0 VALUE DOUBLE?

-1 VALUE DP-LOCATION

\ a version of NUMBER? that detects a decimal point.

: NUMBER?       ( addr len -- d1 f1 )
                FALSE TO DOUBLE?                \ initially not a double #
                -1 TO DP-LOCATION
                OVER C@ [CHAR] - =
                OVER 0> AND DUP>R
                IF      1 /STRING
                THEN
                DUP 0=
                IF      R>DROP
                        2DROP 0 0 FALSE _EXIT   \ always return zero on failure
                THEN
                0 0 2SWAP >NUMBER
                OVER C@ [CHAR] . =              \ next char is a '.'
                OVER 0> AND                     \ more chars to look at
                IF      DUP 1- TO DP-LOCATION
                        1 /STRING >NUMBER
                        DUP 0=
                        IF      TRUE TO DOUBLE? \ mark as a double number
                        THEN
                THEN    NIP 0=
                R>
                IF      >R DNEGATE R>
                THEN    ;

: ?MISSING      ( f -- )
                ABORT" is undefined" ;

: (NUMBER)      ( str -- d )
                COUNT FIND-BUFFER PLACE
                FIND-BUFFER ?UPPERCASE COUNT NUMBER? 0= ?MISSING ;

DEFER NUMBER   ' (NUMBER) IS NUMBER

\ -------------------- Number Output --------------------

: SPACE         ( -- )
                BL EMIT ;

128 CONSTANT SPCS-MAX  ( optimization for SPACES )

CREATE SPCS         SPCS-MAX ALLOT
       SPCS THERE   SPCS-MAX BLANK

: SPACES        ( n -- )
                BEGIN   DUP 0>
                WHILE   DUP SPCS-MAX MIN
                        SPCS OVER TYPE
                        -
                REPEAT  DROP ;

: _COL          ( n -- )
                _GETCOLROW DROP 1- MIN _GETXY DROP - SPACES ;

DEFER COL       ' _COL       IS COL

: #TAB          ( n1 -- )
                GETXY DROP OVER / 1+ * COL ;

: .ID           ( nfa -- )
                NFA-COUNT TYPE SPACE ;

: DECIMAL       10 BASE ! ;
: HEX           16 BASE ! ;
: BINARY         2 BASE ! ;
: OCTAL          8 BASE ! ;

CODE HOLD       ( char -- )
                mov     eax, HLD [UP]    
                dec     eax
                mov     0 [eax] [edi], bl
                mov     HLD [UP]  , eax
                pop     ebx
                next    c;

: <#            ( -- )
                PAD HLD ! ;

: #>            ( d1 -- addr len )
                2DROP  HLD @ PAD OVER - ;

: SIGN          ( f1 -- )
                0< IF  [CHAR] - HOLD  THEN ;

CODE #          ( d1 -- d2 )
                push    edx                     \ save UP
                mov     ecx, BASE [UP]    
                sub     edx, edx
                mov     eax, ebx
                div     ecx
                mov     ebx, eax
                mov     eax, 4 [esp]
                div     ecx
                mov     4 [esp], eax
                mov     eax, edx
                pop     edx                     \ restore UP
                cmp     al, # 9
                jbe     short @@1
                add     al, # 7
@@1:            add     al, # CHAR 0
                mov     ecx, HLD [UP]    
                dec     ecx
                mov     0 [ecx] [edi], al
                mov     HLD [UP]  , ecx
                next    c;

: #S            ( d1 -- d2 )            BEGIN  #  2DUP OR 0= UNTIL ;
: (D.)          ( d -- addr len )       TUCK DABS  <# #S ROT SIGN #> ;
: D.            ( d -- )                (D.) TYPE SPACE ;
: D.R           ( d w -- )              >R (D.) R> OVER - SPACES TYPE ;
: .             ( n -- )                S>D  D. ;
: .R            ( n w -- )              >R  S>D  R>  D.R ;
: U.            ( u -- )                0 D. ;
: U.R           ( u w -- )              0 SWAP D.R ;
: H.            ( u -- )                BASE @ SWAP  HEX U.  BASE ! ;
: ?             ( addr -- )             @ . ;

\ -------------------- Parse Input Stream --------------------

: SOURCE        ( -- addr len )   (SOURCE) 2@ ;

CODE WORD       ( char -- addr )
                push    esi
                push    edi
                mov     al, bl                  \ al = delimiter
                mov     ebx, edi                \ ebx is now our base pointer
                mov     edi, S_ADR [ebx]        \ edi = input pointer
                add     edi, >IN [ebx]          \ add >in
                add     edi, ebx                \ make absolute address
                mov     ecx, S_LEN [ebx]        \ ecx = input length
                sub     ecx, >IN [ebx]          \ subtract >in
                ja      short @@9
                sub     ecx, ecx                \ at end of input
                jmp     short @@8
@@9:            cmp     al, # 32
                jne     short @@5
        \ Delimiter is a blank, treat all chars <= 32 as the delimiter
@@1:            cmp     0 [edi], al             \ leading delimiter?
                ja      short @@2
                inc     edi                     \ go to next character
                dec     ecx
                jnz     @@1
                mov     esi, edi                \ esi = start of word
                mov     ecx, edi                \ ecx = end of word
                jmp     short @@7
@@2:            mov     esi, edi                \ esi = start of word
@@3:            cmp     0 [edi], al             \ end of word?
                jbe     short @@4
                inc     edi
                dec     ecx
                jnz     @@3
                mov     ecx, edi                \ ecx = end of word
                jmp     short @@7
@@4:            mov     ecx, edi                \ ecx = end of word
                inc     edi                     \ skip over ending delimiter
                jmp     short @@7
        \ delimiter is not a blank
@@5:            repz    scasb
                jne     short @@6
                mov     esi, edi                \ end of input
                mov     ecx, edi
                jmp     short @@7
@@6:            dec     edi                     \ backup
                inc     ecx
                mov     esi, edi                \ esi = start of word
                repnz   scasb
                mov     ecx, edi                \ ecx = end of word
                jne     short @@7
                dec     ecx                     \ account for ending delimiter
        \ Update >IN pointer and get word length
@@7:            sub     edi, ebx                \ relative address
                sub     edi, S_ADR [ebx]        \ offset from start
                mov     >IN [ebx], edi          \ update >IN
                sub     ecx, esi                \ length of word
                cmp     ecx, # MAXCOUNTED       \ max at MAXCOUNTED
                jbe     short @@8
                mov     ecx, # MAXCOUNTED       \ clip to MAXCOUNTED
        \ Move string to pocket
@@8:            lea     edi, POCKET [ebx]       \ edi = pocket
                mov     0 [edi], cl             \ store count byte
                inc     edi
                rep     movsb                   \ move rest of word
                sub     eax, eax                \ clear EAX
                stosb                           \ append a NULL to pocket
                mov     ebx, # pocket           \ return pocket
                pop     edi
                pop     esi
                next    c;

: PARSE         ( char -- addr len )
                >R SOURCE >IN @ /STRING
                2DUP R> SCAN NIP -
                DUP 1+ >IN +! ;

: .(            ( -- )
                [CHAR] ) PARSE TYPE ; IMMEDIATE

: \             ( -- )
                SOURCE >IN ! DROP ; IMMEDIATE

: <A            SOURCE >IN ! DROP ; IMMEDIATE   \ HTML link start marker

\ -------------------------------------------------------
\ Header structure as of June 1996
\
\       [ optimize field   ]              OFA
\       [ alignment bytes  ] 0 to 3 bytes for header alignment
\       [ the name letters ]
\       [ count byte       ] -9  -5       NFA
\       [ view field       ] -8  -4       VFA
\       [ link field       ] -4  +0       LFA
\       [ cfa ptr field    ] +0           CFA-PTR
\               |          
\               V          
\       [ cfa  field       ] +0           CFA
\       [ body field       ] +4           PFA
\ -------------------------------------------------------


\ -------------------------------------------------------
\ Vocabulary dictionary structure
\
\       [ cfa field        ] +0           VCFA = vocabulary cfa
\       [ voc link         ] +4           VLINK
\       [ num voc threads  ] +8           #THREADS
\       [ voc thread 0     ] +12          VOC thread 0 = voc-address
\       [ voc thread 1     ] +16          VOC thread 1
\       [ voc thread 2     ] +20          VOC thread 2
\       [ voc thread ...   ] +n*4+12      VOC thread n
\ -------------------------------------------------------

CODE >BODY      ( cfa -- pfa )
                add     ebx, # 4        \ cell+
                next    c;

CODE BODY>      ( pfa -- cfa )
                sub     ebx, # 4        \ cell-
                next    c;

CODE L>NAME     ( lfa -- nfa )
                sub     ebx, # 5        \ 1- cell-
                next    c;

CODE LINK>      ( link -- cfa )
                add     ebx, # 4        \ cell+ @
                mov     ebx, 0 [ebx] [edi]
                next    c;

CODE N>LINK     ( nfa -- lfa )
                add     ebx, # 5        \ cell+ 1+
                next    c;

CODE VLINK>VOC  ( voc-link-field -- voc-address )
                add     ebx, # 4        \ cell+
                next    c;

CODE VOC>VLINK  ( voc-address -- voc-link-field )
                sub     ebx, # 4        \ cell -
                next    c;

CODE VOC#THREADS ( voc-address -- #threads )
                sub     ebx, # 8        \ 2 cells - @
                mov     ebx, 0 [ebx] [edi]
                next    c;

CODE VCFA>VOC   ( vocabulary-cfa -- voc-address )
                add     ebx, # 12       \ >BODY 2 CELLS+
                next    c;

CODE VOC>VCFA   ( voc-address -- vocabulary-cfa )
                sub     ebx, # 12       \ 2 CELLS - BODY>
                next    c;

VARIABLE [UNKNOWN]

\ search thread for cfa contents of [UNKNOWN], return its nfa if found

CODE ?THREAD    ( FALSE thread -- FALSE FALSE | nfa TRUE TRUE )
@@2:            mov     ebx, 0 [ebx] [edi]        \ get contents of thread
                test    ebx, ebx                \ if top of stack = zero
                je      short @@1               \ branch to exit
                mov     eax, 4 [ebx] [edi]      \ get the CFA
                sub     eax, [unknown] [edi]    \ does CFA match?
                test    eax, eax
                jne     short @@2               \ if they match
                sub     ebx, # 5
                pop     eax
                push    ebx
                mov     ebx, # TRUE
                push    ebx
                jmp     short @@3
@@1:            mov     ebx, # FALSE            \ extra FALSE on stack
@@3:            next    c;

: _>NAME        ( cfa -- nfa )  \ find the nfa of the provided cfa.
                                \ returns the nfa of [UNKNOWN] if the nfa of
                                \ cfa can't be found
                BEGIN   [UNKNOWN] !
                        VOC-LINK
                        BEGIN   @ ?DUP
                        WHILE   FALSE
                                OVER VLINK>VOC DUP VOC#THREADS CELLS BOUNDS
                                ?DO     I ?THREAD ?LEAVE
\                                       BEGIN   @ ?DUP
\                                       WHILE   DUP LINK> [UNKNOWN] @ = \ cfa's match?
\                                               IF      L>NAME NIP
\                                                       TRUE LEAVE
\                                               THEN
\                                       REPEAT
                          CELL +LOOP
                                IF      NIP
                                        _EXIT           \ LEAVE WE FOUND IT
                                THEN
                        REPEAT  ['] [UNKNOWN]
                AGAIN   ;

DEFER >NAME     ' _>NAME IS >NAME

CODE NAME>      ( nfa -- cfa )
                add     ebx, # 9                \ 9 + @
                mov     ebx, 0 [ebx] [edi]
                next    c;

CODE N>CFAPTR   ( nfa -- cfa-pointer )
                add     ebx, # 9                \ 9 +
                next    c;

: >VIEW         ( cfa -- vfa )
                >NAME 1+ ;

CODE VIEW>      ( vfa -- cfa )
                add     ebx, # 8                \ 8 + @
                mov     ebx, 0 [ebx] [edi]
                next    c;

: N>OFA         ( nfa -- ofa )
                DUP
                IF      NFA-COUNT DROP -ALIGNED CELL-
                THEN    ;

: >OFA          ( cfa -- ofa )          \ get the Optimization Field Address
                >NAME N>OFA ;

\ -------------------- Dictionary Search --------------------

VARIABLE LATEST-NFA

CODE SEARCH-1WORDLIST    ( addr len wid -- 0 | cfa flag )
                pop     eax                     \ eax = length
                xchg    ebp, esp                \ [ebp] = addr
                push    esi
                push    edi
                push    edx                     \ save UP
                test    eax, eax                \ zero length does not match
                je      short @@5
                mov     edx, edi                \ edx = new base pointer
                add     0 [ebp], edx            \ make string address absolute
                jmp     short @@2
@@1:            mov     cl, -5 [ebx] [edx]      \ count byte
                and     ecx, # 63               \ 05/13/94 tjz max length now 63
                cmp     ecx, eax                \ counts equal?
                je      short @@3
@@2:            mov     ebx, 0 [ebx] [edx]      \ follow link
                test    ebx, ebx
                jne     short @@1               \ end of chain?
@@5:            add     ebp, # 4                \ discard string address
                sub     ebx, ebx                \ return not found
                jmp     short @@4
@@3:            mov     esi, 0 [ebp]              \ esi = string address
                lea     edi, -5 [ebx] [edx]     \ edi = name field
                mov     LATEST-NFA [edx], edi   \ save nfa for other use
                sub     LATEST-NFA [edx], edx   \ convert from abs to rel
                sub     edi, ecx                \ edi = first char of name
                repz    cmpsb
                jne     short @@2
                mov     eax, 4 [ebx] [edx]      \ cfa
                mov     0 [ebp], eax            \ this is the cfa
                mov     al, -5 [ebx] [edx]      \ get count byte
                mov     ebx, # 1                \ assume word is immediate
                and     al, # 128               \ test immediate bit
                jne     short @@4
                neg     ebx                     \ not immediate
@@4:            pop     edx                     \ restore UP & exit
                pop     edi
                pop     esi
                xchg    ebp, esp
                next    c;

CODE "#HASH     ( a1 n1 #threads -- n2 )
                pop     eax                     \ pop count into EAX
                mov     ecx, ebx                \ save #threads in ecx
                pop     ebx                     \ get string address into EBX
                push    edx                     \ save UP
                xor     edx, edx                \ clear EDX
                cmp     eax, # 0                \ if count is zero
                je      @@2                     \ then exit
                push    ecx                     \ save #threads for later
                mov     ecx, eax                \ copy count into ECX
@@1:            rol     eax, # 7                \ rotate result some
                mov     dl, 0 [ebx] [edi]         \ get first character of name
                add     eax, edx                \ accumulate another character
                inc     ebx                     \ bump to address of next    c; character
                dec     ecx
                jnz     @@1
                xor     edx, edx                \ clear high part of dividend
                pop     ebx                     \ get number of threads
                div     ebx                     \ perform modulus
@@2:            mov     ebx, edx                \ move result into EBX
                lea     ebx, 0 [ebx*4]            \ multiply by cell size
                pop     edx                     \ restore UP
                next    c;

: _SEARCH-WORDLIST ( addr len voc -- 0 | cfa flag )
                >R 2DUP R@ VOC#THREADS "#HASH R> + SEARCH-1WORDLIST ;

63 VALUE NAME-MAX-CHARS                 \ function names can be this long 

: SEARCH-WORDLIST ( addr len voc -- 0 | cfa flag )
                >R FIND-BUFFER PLACE
                FIND-BUFFER ?UPPERCASE COUNT
                2DUP R@ VOC#THREADS "#HASH R> + SEARCH-1WORDLIST ;

\ WARNING: (FIND) is a case sensitive find.  If you need to be able to find
\ words in the dictionary that have not already been passed through ?UPPERCASE,
\ then you should use CAPS-FIND which will uppercase the string before trying
\ to find it in the dictionary.

: (FIND)        ( str -- str FALSE | cfa flag )
                DUP C@ 0= IF 0 _EXIT THEN
                CONTEXT
                BEGIN   DUP @                   \ while not at end of list
                WHILE   DUP 2@ <>               \ and not the same vocabulary
                                                \ as NEXT time
                        IF      OVER COUNT NAME-MAX-CHARS MIN
                                2 PICK @ _SEARCH-WORDLIST ?DUP
                                IF      2SWAP 2DROP     \ found it, so
                                        _EXIT           \ we're done searching
                                THEN
                        THEN    CELL+                   \ step to next vocabulary
                REPEAT  DROP
                FALSE  ;

: CAPS-FIND     ( str -- str FALSE | cfa flag )
                DUP COUNT FIND-BUFFER PLACE
                FIND-BUFFER ?UPPERCASE (FIND)
                DUP>R
                IF      NIP
                ELSE    DROP
                THEN     R> ;

DEFER FIND      ( str -- str 0 | cfa flag )

: DEFINED       ( -- str 0 | cfa flag )
                BL WORD PARMFIND ;

\ -------------------- Compiler --------------------

CODE APP-HERE   ( -- a1 ) \ return a1 the address of the next free byte of
                          \ application dictionary space
                push    ebx
                mov     ebx, DP [edi]
                next    c;

DEFER HERE      ' APP-HERE IS HERE

CODE SYS-HERE   ( -- a1 ) \ return a1 the address of the next free byte of
                          \ system dictionary space
                push    ebx
                mov     ebx, SDP [edi]
                next    c;

CODE APP-ALLOT  ( n1 -- ) \ allot n1 bytes in the application dictionary
                add     DP [edi], ebx
                pop     ebx
                next    c;

DEFER VEC-ALLOT ' APP-ALLOT IS VEC-ALLOT

CODE SYS-ALLOT  ( n1 -- ) \ allot n1 bytes in the system dictionary
                add     SDP [edi], ebx
                pop     ebx
                next    c;

: SYS-ORIGIN    ( -- a1 ) \ return the absolute origin of system dictionary
                &ORIGIN @ &SYS-SEPARATION @ + ;

: ?MEMCHK       ( n1 -- ) \ test to see if we have enough memory
                0MAX DUP &ORIGIN &APP-SIZE @ + U>
                SWAP 512 + APP-HERE + &ORIGIN &APP-SIZE @ + U> OR
                ABORT" OUT OF APPLICATION MEMORY!"
                &SYS-SIZE @ SYS-HERE SYS-ORIGIN - - 512 <
                ABORT" Out of System/Header memory!" ;

: ALLOT         ( n1 -- ) \ define ALLOT with a memory full check
                DUP 1000 + ?MEMCHK VEC-ALLOT ;

CODE _APP-COMPILE, ( xt -- )
                mov     eax, DP [edi]
                mov     0 [eax] [edi], ebx
                add     eax, # 4
                mov     DP [edi], eax
                pop     ebx
                next    c;

TRUE VALUE SYS-WARNING?

: ?SYS-ADDRESS  ( a1 -- ) \ check that a1 is not in system dictionary,
                          \ warn user if it is.
                DUP APP-HERE U>                 \ in system space
                SYS-WARNING? AND                \ and we want warnings
                IF      BASE @ >R DECIMAL
                        CR ." ****System word: "
                        DUP SYS-HERE U<
                        IF      DUP .NAME
                        ELSE    DUP H.
                        THEN    ." used in: "   LAST @ NFA-COUNT TYPE
                        cr ."  at file: " LOADFILE @ CELL+ COUNT TYPE
                        ."  line: " LOADLINE @ .
                        R> BASE !
                THEN    DROP ;

: APP-COMPILE,  ( xt -- )
                DUP ?SYS-ADDRESS _APP-COMPILE, ;

DEFER COMPILE,  ' APP-COMPILE, IS COMPILE,

: SYS-WARNING-OFF ( -- ) \ disable warning for use of system words in application
                FALSE TO SYS-WARNING? ;

: SYS-WARNING-ON  ( -- ) \ enable warning for use of system words in application
                TRUE  TO SYS-WARNING? ;

CODE SYS-COMPILE, ( xt -- )
                mov     eax, SDP [edi]
                mov     0 [eax] [edi], ebx
                add     eax, # 4
                mov     SDP [edi], eax
                pop     ebx
                next    c;

CODE APP-,      ( n1 -- )
                mov     eax, DP [edi]
                mov     0 [eax] [edi], ebx
                add     eax, # 4
                mov     DP [edi], eax
                pop     ebx
                next    c;

DEFER ,         ' APP-, IS ,

CODE SYS-,      ( n1 -- )
                mov     eax, SDP [edi]
                mov     0 [eax] [edi], ebx
                add     eax, # 4
                mov     SDP [edi], eax
                pop     ebx
                next    c;

CODE APP-C,     ( c1 -- )
                mov     eax, DP [edi]
                mov     0 [eax] [edi], bl
                inc     eax
                mov     DP [edi], eax
                pop     ebx
                next    c;

DEFER C,        ' APP-C, IS C,

CODE SYS-C,     ( c1 -- )
                mov     eax, SDP [edi]
                mov     0 [eax] [edi], bl
                inc     eax
                mov     SDP [edi], eax
                pop     ebx
                next    c;

CODE APP-W,     ( w1 -- )
                mov     eax, DP [edi]
                16:
                mov     0 [eax] [edi], ebx
                add     eax, # 2
                mov     DP [edi], eax
                pop     ebx
                next    c;

DEFER W,        ' APP-W, IS W,

CODE SYS-W,     ( w1 -- )
                mov     eax, SDP [edi]
                16:
                mov     0 [eax] [edi], ebx
                add     eax, # 2
                mov     SDP [edi], eax
                pop     ebx
                next    c;

CODE APP-COMPILE  ( -- )
                lodsd
                mov     ecx, DP [edi]
                mov     0 [ecx] [edi], eax
                add     ecx, # 4
                mov     DP [edi], ecx
                next    c;

DEFER COMPILE   ' APP-COMPILE IS COMPILE

CODE SYS-COMPILE ( -- )
                lodsd
                mov     ecx, SDP [edi]
                mov     0 [ecx] [edi], eax
                add     ecx, # 4
                mov     SDP [edi], ecx
                next    c;

CODE APP-ALIGN  ( -- )
                mov     eax, DP [edi]
                jmp     short @@2
@@1:            mov     cl, # 0
                mov     0 [eax] [edi], cl
                inc     eax
@@2:            mov     ecx, eax
                and     ecx, # 3
                jne     short @@1
                mov     DP [edi], eax
                next    c;

DEFER ALIGN     ' APP-ALIGN IS ALIGN

CODE SYS-ALIGN  ( -- )
                mov     eax, SDP [edi]
                jmp     short @@2
@@1:            mov     cl, # 0
                mov     0 [eax] [edi], cl
                inc     eax
@@2:            mov     ecx, eax
                and     ecx, # 3
                jne     short @@1
                mov     SDP [edi], eax
                next    c;

CODE ALIGNED    ( addr1 -- addr2 )
                add     ebx, # 3
                and     ebx, # -4
                next    c;

CODE -ALIGNED   ( addr1 -- addr2 )
                and     ebx, # -4
                next    c;

FALSE VALUE IN-SYSTEM?

: IN-APPLICATION ( -- )
                FALSE TO IN-SYSTEM?
                ['] APP-HERE     IS HERE
                ['] APP-ALLOT    IS VEC-ALLOT
                ['] APP-COMPILE, IS COMPILE,
                ['] APP-,        IS ,
                ['] APP-C,       IS C,
                ['] APP-W,       IS W,
                ['] APP-COMPILE  IS COMPILE
                ['] APP-ALIGN    IS ALIGN ;

: IN-SYSTEM     ( -- )
                TRUE TO IN-SYSTEM?
                ['] SYS-HERE     IS HERE
                ['] SYS-ALLOT    IS VEC-ALLOT
                ['] SYS-COMPILE, IS COMPILE,
                ['] SYS-,        IS ,
                ['] SYS-C,       IS C,
                ['] SYS-W,       IS W,
                ['] SYS-COMPILE  IS COMPILE
                ['] SYS-ALIGN    IS ALIGN ;

: >SYSTEM       ( -- )          \ select system dictionary, save prev dict
                R> IN-SYSTEM? >R >R IN-SYSTEM ;

: SYSTEM>       ( -- )  \ return to previous dictionary
                R> R> 0=
                IF      IN-APPLICATION
                THEN    >R ;

: >APPLICATION  ( -- )  \ select system dictionary, save prev dict
                R> IN-SYSTEM? >R >R IN-APPLICATION ;

: APPLICATION>  ( -- )  \ return to previous dictionary
                R> R>
                IF      IN-SYSTEM
                THEN    >R ;

: IMMEDIATE     ( -- )  \ mark the last header create as an immediate word
                LAST @ 128 TOGGLE ;

: HIDE          ( -- )
                LAST @ N>LINK @ LAST-LINK @ ! ;

: REVEAL        ( -- )
                LAST @ N>LINK LAST-LINK @ ! ;

: LITERAL       ( n -- )
                COMPILE LIT , ; IMMEDIATE

: CHAR          ( -- char )
                BL WORD 1+ C@ ;

: [CHAR]        ( -- char )
                CHAR  [COMPILE] LITERAL ; IMMEDIATE

: '             ( -- cfa )
                DEFINED 0= ?MISSING ;

: [']           ( -<name>- )
                ' [COMPILE] LITERAL ; IMMEDIATE

: [COMPILE]     ( -<name>- )
                ' COMPILE, ; IMMEDIATE

: POSTPONE      ( -<name>- )
                DEFINED DUP 0= ?MISSING
                0< IF  COMPILE COMPILE  THEN
                COMPILE, ; IMMEDIATE

\ -------------------- String Literals --------------------

\ Convert occurances of \N within string a1,n1 to the CRLF pairs.
\ Useful mostly for strings that will get passed to the operating system.

DEFER \N->CRLF  ( a1 n1 -- )    ' 2DROP IS \N->CRLF

: ,"            ( -<string">- )
                [CHAR] " PARSE
                HERE >R DUP C,
                DUP ALLOT
                R@ 1+ SWAP MOVE
                0 C,
                ALIGN
                R> COUNT \N->CRLF ;

NCODE (("))     ( -- counted-string )
                push    ebx
                mov     eax, 0 [ebp]
                mov     ebx, eax
                sub     ebx, edi        \ relative address
                sub     ecx, ecx
                mov     cl, 0 [eax]
                add     eax, ecx
                add     eax, # 5        \ account for extra null after string
                and     eax, # -4       \ align
                mov     0 [ebp], eax
                next    c;

DEFER NEW$    ' TEMP$ IS NEW$           ( a1 -- a2 )

: (C")          ( -- counted-string )
                ((")) ;

: C"            ( -<string">- )
                STATE @
                IF      COMPILE (C")  ,"
                ELSE    [CHAR] " WORD
                        NEW$ DUP>R OVER C@ 1+ MOVE
                        r>
                THEN ; IMMEDIATE

: (S")          ( -- addr len )
                ((")) COUNT ;

: S"            ( -<string">- )
                STATE @
                IF      COMPILE (S")  ,"
                ELSE    [CHAR] " WORD
                        NEW$ DUP>R OVER C@ 1+ MOVE
                        r> COUNT
                THEN ; IMMEDIATE

: (.")          ( -- )
                ((")) COUNT TYPE ;

: ."            ( -<string">- )
                COMPILE (.")  ,"  ; IMMEDIATE

\ -------------------- Error Handler --------------------

: CATCH         ( cfa -- flag )
                SP@ >R
                LP @ >R
                OP @ >R
                HANDLER @ >R
                RP@ HANDLER !
                EXECUTE
                R> HANDLER !
                R> DROP
                R> DROP
                R> DROP
                0 ;

: THROW         ( n -- )
                ?DUP
                IF      HANDLER @ RP!
                        R> HANDLER !
                        R> OP !
                        R> LP !
                        R> SWAP >R
                        SP! DROP
                        R>
                THEN ;

: ABORT         ( -- )
                THROW_ABORT THROW ;

: (ABORT")      ( f -- )
                ((")) SWAP
                IF      MSG !
                        THROW_ABORTQ THROW
                THEN    DROP ;

: ABORT"        ( -- )
                COMPILE (ABORT")  ,"  ; IMMEDIATE

\ -------------------- Structured Conditionals --------------------

: ?EXEC  STATE @    ABORT" execution only"    ;
: ?COMP  STATE @ 0= ABORT" compilation only"  ;

: ?PAIRS        ( n1 n2 -- )  XOR ABORT" conditionals not paired"  ;

: >MARK         ( -- addr )   HERE 0 , ;
: >RESOLVE      ( addr -- )   HERE OVER -  SWAP ! ;
: <MARK         ( -- addr )   HERE ;
: <RESOLVE      ( addr -- )   HERE - , ;

: AHEAD  ?COMP  COMPILE  BRANCH  >MARK 2 ; IMMEDIATE
: IF     ?COMP  COMPILE ?BRANCH  >MARK 2 ; IMMEDIATE
: THEN   ?COMP  2 ?PAIRS  COMPILE _THEN  >RESOLVE ; IMMEDIATE
: ENDIF  [COMPILE] THEN ; IMMEDIATE
: ELSE   ?COMP  2 ?PAIRS  COMPILE BRANCH >MARK  SWAP >RESOLVE  2 ; IMMEDIATE

: BEGIN  ?COMP  COMPILE _BEGIN  <MARK 1 ; IMMEDIATE
: UNTIL  ?COMP  1 ?PAIRS  COMPILE _UNTIL  <RESOLVE ; IMMEDIATE
: AGAIN  ?COMP  1 ?PAIRS  COMPILE _AGAIN  <RESOLVE ; IMMEDIATE
: WHILE  ?COMP  COMPILE _WHILE  >MARK 2  2SWAP ; IMMEDIATE
: REPEAT ?COMP  1 ?PAIRS  COMPILE _REPEAT <RESOLVE  2 ?PAIRS >RESOLVE ; IMMEDIATE

: DO     ?COMP  COMPILE (DO)   >MARK 3 ; IMMEDIATE
: ?DO    ?COMP  COMPILE (?DO)  >MARK 3 ; IMMEDIATE
: LOOP   ?COMP  3 ?PAIRS  COMPILE (LOOP)   DUP CELL+ <RESOLVE  >RESOLVE ; IMMEDIATE
: +LOOP  ?COMP  3 ?PAIRS  COMPILE (+LOOP)  DUP CELL+ <RESOLVE  >RESOLVE ; IMMEDIATE

\ -------------------- Eaker CASE statement --------------------

: CASE   ?COMP  COMPILE _CASE  0 ; IMMEDIATE
: OF     ?COMP  COMPILE _OF  >MARK 4 ; IMMEDIATE
: ENDOF  ?COMP  4 ?PAIRS  COMPILE _ENDOF  >MARK  SWAP >RESOLVE  5 ; IMMEDIATE

: ENDCASE  ?COMP  COMPILE _ENDCASE
           BEGIN  ?DUP WHILE  5 ?PAIRS  >RESOLVE  REPEAT ; IMMEDIATE

\ -------------------- Build Header --------------------

: LINK,         ( addr -- )    \ build link from list head at addr
                HERE  OVER @ ,  SWAP !  ;

: SYS-LINK,     ( addr -- )     \ build link from list head at addr
                SYS-HERE  OVER @ SYS-,  SWAP !  ;

: VIEW,         ( -- )          \ compile the view field
                ?LOADING @
                IF      LOADLINE @ SYS-,
                ELSE    -1 SYS-,
                THEN    ;

0 VALUE OFA-LAST

: "NAME,        ( a1 n1 -- )    \ align and compile name a1,n1 at here
                NAME-MAX-CHARS MIN
                SYS-ALIGN               \ align system space
                APP-ALIGN               \ align application space
                DUP 0= ABORT" Need a NAME to create!"
                SYS-HERE TO OFA-LAST
                -1 SYS-,                \ OFA cell for optimizing compiler
                2>R
                CAPS @
                IF      2R@ UPPER
                THEN
                3 2R@ NIP 3 AND - SYS-ALLOT \ pre-align for name length
                2R@ CURRENT @ _SEARCH-WORDLIST
                IF      WARNING @
                        IF      CR ?LOADING @
                                IF      ." From file: " CUR-FILE COUNT TYPE
                                        ."  word: "
                                THEN    2R@ TYPE ."  isn't unique "
                        THEN    DROP
                THEN
                2R> >R SYS-HERE R@ MOVE
                R@ SYS-ALLOT
                SYS-HERE LAST !
                R> SYS-C, ;

: _"HEADER      ( a1 n1 -- )    \ build a hashed header from a1,n1
                NAME-MAX-CHARS MIN 2DUP 2>R "NAME, VIEW,
                CURRENT @ DUP 2R> ROT VOC#THREADS "#HASH +
                DUP LAST-LINK ! SYS-LINK,
\ the following line looks obscure, but what it does is make sure the CFA
\ follows the CFA-pointer when compiling definitions into system space
                SYS-HERE 0 SYS-, HERE SWAP ! ;

DEFER "HEADER   ' _"HEADER IS "HEADER

: _HEADER       ( -<name>- )    \ build a header, but check available memory
                2000 ?MEMCHK
                BL WORD COUNT "HEADER ;

DEFER HEADER    ( -<name>- )
                ' _HEADER IS HEADER

: CREATE        ( -<name>- )
                HEADER  DOVAR ,  ;

: ALIAS         ( xt -<name>- ) \ make another 'name' for 'xt'
                HEADER LAST @ N>CFAPTR ! ;
                                \ Useage is  ' OLD ALIAS NEW
                                \ See also SYNONYM for an immediate version                    

\ -------------------- Colon Compiler --------------------

VARIABLE CSP    \ Current Stack Pointer variable

: !CSP          ( -- )  \ save current stack pointer for later stack depth check
                SP@ CSP ! ;

: ?CSP          ( -- )  \ check current stack pointer against saved stack pointer
                SP@ CSP @ XOR ABORT" stack changed" ;

: CALL,         ( addr -- )  \ compile call
                [ HEX E8909090 DECIMAL ] LITERAL ,
                HERE CELL+ - , ;

: (;CODE)       ( -- )
                R> ABS>REL  LAST @ NAME> !  ;

: #(;CODE)      ( a1 -- )
                R> ABS>REL SWAP ! ;

: _]            ( -- )
                STATE ON ;

: _[            ( -- )          \ turn off compiling
                STATE OFF ;

DEFER ]           ' _] IS ]
DEFER [ IMMEDIATE ' _[ IS [     \ turn off compiling

: :NONAME       ( -- xt )       \ start a headerless colon definition
                ALIGN HERE   DOCOL ,  !CSP ] ;

0 VALUE ?:M
0 VALUE PARMS   \ number of parameters

: PARMS-INIT    ( -- )
                FALSE TO ?:M
                0 TO PARMS ;

: _:            ( -<name>- )    \ Forth's primary function defining word
                PARMS-INIT
                HEADER HIDE  DOCOL ,  !CSP ] ;

DEFER :   ' _: IS :

: RECURSE       ( -- )          \ cause current definition to execute itself
                ?COMP  LAST @ NAME> COMPILE, ; IMMEDIATE

\ -------------------- Defining Words --------------------

: CONSTANT      ( n -<name>- )  \ create a constant (unchangable) value
                HEADER  DOCON ,  ,  ;

: VARIABLE      ( -<name>- )    \ create a variable (changable) value
                CREATE 0 , ;

: USER          ( n -<name>- )  \ create a user variable (changable) value
                HEADER  DOUSER ,  , ;

: DEFER         ( -<name>- )    \ create a deferred execution function
                HEADER DODEFER ,
                COMPILE NOOP
                DEFER-LIST LINK,
                COMPILE NOOP ;

: NEWUSER       ( size - )      \ Creates a user. A user can be
                                \ a byte, cell, float, string or stack
                NEXT-USER @ SWAP OVER + NEXT-USER !
                USER ;

: .USERSIZE     ( - )           \ Shows what is left in the user-area
                USERSIZE 3 CELLS - NEXT-USER @
                cr ." Next-user at " DUP .
                cr ." Free user space " - . ." bytes." ;

\ DO VOCABULARY

LABEL DOVOC     ( -- )                  \ "runtime" for VOCABULARY
                add     eax, # 3 CELLS  \ set CONTEXT to VOC address
                mov     context [edi], eax
                next    c;

ASSEMBLER DOVOC META RESOLVES <VOCABULARY>

: DEFINITIONS   ( -- )
                CONTEXT @ CURRENT ! ;

: 2CONSTANT     ( n1 n2 -- )
                CREATE , ,
                ;CODE   0 OFA-H !       \ disable OFA resolution
                        push    ebx
                        push         2 CELLS [eax] [edi]
                        mov     ebx, 1 CELLS [eax] [edi]
                        next    c;
                DROP                    \ cleanup the stack for meta compiler

: 2VARIABLE     ( -<name>- )
                VARIABLE 0 , ;

\ -------------------- Redefine DEFER-red Words --------------------

NCODE @(IP)     ( -- n )
                push    ebx
                mov     eax, 0 [ebp]
                mov     ebx, 0 [eax]
                add     eax, # 1 CELLS
                mov     0 [ebp], eax
                next    c;

CODE >IS        ( xt -- addr )
                add     ebx, # 4                  \ return >BODY
                next    c;

: (IS)          ( xt -- )   @(IP) >IS ! ;

: ?IS           ( xt -- xt )                    \ error if not a deferred word
                DUP @ DODEFER <> ABORT" Use IS on DEFER-red words!" ;

: IS            ( xt -<name>- )
                STATE @
                IF      COMPILE (IS)   ' ?IS ,
                ELSE      DUP APP-HERE U> >R      \ if xt is system
                        ' DUP APP-HERE U< R> AND  \ and deferred word isn't
                        IF      OVER ?SYS-ADDRESS \ warn user about problem
                        THEN    ?IS >IS !         \ store new deferred func
                THEN ; IMMEDIATE

\ -------------------- Value --------------------

: VALUE         ( n -<name>- )  \ creat a self fetching value, like a constant
                                \ but changable
                HEADER
                DOVALUE   ,
                ( n )     ,
                DOVALUE!  ,
                DOVALUE+! ,   ;

: ?TO_CHECK     ( xt -- xt_body )
                DUP @ >R
                >BODY DUP CELL+ @ -1 =  \ no special words
                R@ DOCON   = OR         \ no constants
                R@ DOCOL   = OR         \ no colon definitions
                R@ DODOES  = OR         \ no DOES> words
                R@ DOVAR   = OR         \ no variables
                R> DODEFER = OR         \ no deferred words
                ABORT" Use TO and +TO on VALUE type words!" ;

: TO            ( n -<value_name>- )
                ' ?TO_CHECK CELL+
                STATE @ IF COMPILE, ELSE EXECUTE THEN ; IMMEDIATE

: +TO           ( n -<value_name>- )
                ' ?TO_CHECK CELL+ CELL+
                STATE @ IF COMPILE, ELSE EXECUTE THEN ; IMMEDIATE

: &OF           ( -<value_name>- addr )
                ' >BODY
                STATE @ IF  [COMPILE] LITERAL  THEN ; IMMEDIATE

\ -------------------- Interpreter --------------------

CODE DEPTH      ( -- n )        \ return the current data stack depth
                push    ebx
                mov     ebx, SP0 [UP]    
                add     ebx, edi
                sub     ebx, esp
                sar     ebx, # 2  \ shift right two is divide by 4
                next    c;

: ?STACK        ( -- )          \ check the data stack for stack underflow
                DEPTH 0< ABORT" stack underflow"  ;

: QUERY         ( -- )          \ accept a line of input from the user to TIB
                TIB DUP MAXSTRING ACCEPT (SOURCE) 2!
                >IN OFF
                0 TO SOURCE-ID 0 TO SOURCE-POSITION ;

: _NUMBER,      ( d -- )
                DOUBLE? 0= IF DROP THEN
                STATE @
                IF      DOUBLE? IF  SWAP  [COMPILE] LITERAL  THEN
                        [COMPILE] LITERAL
                THEN ;

DEFER NUMBER,           ' _NUMBER, IS NUMBER,
DEFER SAVE-SRC          ' NOOP     IS SAVE-SRC
DEFER ?UNSAVE-SRC       ' NOOP     IS ?UNSAVE-SRC

: _INTERPRET    ( -- )
                BEGIN   BL WORD DUP C@
                WHILE   SAVE-SRC FIND ?DUP
                        IF      STATE @ =
                                IF      COMPILE,
                                ELSE    EXECUTE ?STACK
                                THEN
                        ELSE    NUMBER NUMBER,
                        THEN    ?UNSAVE-SRC
                REPEAT DROP ;

DEFER INTERPRET    ' _INTERPRET IS INTERPRET

\ -------------------- Evaluate --------------------

: EVALUATE      ( addr len -- ) \ interpret string addr,len
                SOURCE 2>R
                >IN @ >R
                SOURCE-ID >R
                SOURCE-POSITION >R
                (SOURCE) 2!
                >IN OFF
                -1 TO SOURCE-ID
                0 TO SOURCE-POSITION
                ['] INTERPRET CATCH
                R> TO SOURCE-POSITION
                R> TO SOURCE-ID
                R> >IN !
                2R> (SOURCE) 2!
                THROW ;

: WINPAUSE      ( -- )  \ release control to OS for a moment
                KEY? DROP ;

\ -------------------- File Loading --------------------

VARIABLE ECHO

: ?.REFILL      ( -- )
                ECHO @
                IF      CR SOURCE TYPE
                        START/STOP
                THEN    ;

DEFER .REFILL   ' ?.REFILL IS .REFILL

\ August 11th, 1997 - 9:02 tjz added to correct for performance bug
0 value len-prev

: REFILL        ( -- f )        \ refill TIB from current input stream
                SOURCE-ID ?DUP
                IF      1+  ( not from evaluate )
                        IF      1 LOADLINE +!
                                LOADLINE @ 255 AND 0=   \ once each 256 lines
                                IF      WINPAUSE        \ release control to OS
                                THEN                    \ for a moment
                                TIB DUP MAXSTRING
\ August 11th, 1997 - 9:02 tjz correction for performance bug
\                                SOURCE-ID FFILE-POSITION DROP ( flag )
\                                DROP ( high part ) TO SOURCE-POSITION
        len-prev +to source-position
                                SOURCE-ID FREAD-LINE ABORT" read error"
                                IF      dup 2 + to len-prev
                                        (SOURCE) 2!
                                        >IN OFF
                                        .REFILL
                                        TRUE _EXIT
                                else    0 to len-prev 
                                THEN
                                2DROP
                        THEN
                        FALSE _EXIT
                THEN
                CR QUERY TRUE ;

: LINKFILE      ( a1 -- ) \ link name a1 as current file IF LOADING ONLY !!
                ?LOADING @
                IF      LOADFILE SYS-LINK,
                        COUNT SYS-HERE PLACE
                        SYS-HERE C@ 1+ SYS-ALLOT
                ELSE    DROP
                THEN ;

DEFER STACK-CHECK       ' NOOP   IS STACK-CHECK

VARIABLE START-LINE

FALSE VALUE INCLUDING?

: >LINE         ( n1 -- )     \ move to line n1, 1 based
                1- 0 MAX
                ?DUP
                IF      0 DO  REFILL DROP  LOOP
                THEN ;

: DO-INCLUDE    ( -- )
                INCLUDING? >R
                TRUE TO INCLUDING?
                START-LINE @ >LINE
                START-LINE OFF
\ August 11th, 1997 - 9:02 tjz two lines added to correct for performance bug
                source-position >r 0 to source-position
                len-prev        >r 0 to len-prev
                BEGIN   REFILL
                        INCLUDING? AND
                WHILE   INTERPRET
                        STACK-CHECK
                REPEAT
\ August  11th, 1997 -  9:02 tjz two lines added to correct for performance bug
\ October 24th, 1997 - 11:47 tjz reversed following two lines to correct bug
                r> to len-prev
                r> to source-position
                R> TO INCLUDING? ;

16 CONSTANT DEFEXTMAX
CREATE DEFEXT$  ( -- a1 ) \ the default extension buffer
                ," F" 0 , 0 , 0 , 0 ,

VARIABLE EXT_ADDED?             \ January 17th, 2000 - 10:31 added to better
                                \ handle files without extensions
   FALSE EXT_ADDED? !-T

VARIABLE DEFEXT_ON?             \ January 17th, 2000 - 10:31 added to better
                                \ handle files without extensions
    TRUE DEFEXT_ON? !-T

: DEFEXT        ( -<FTH>- ) \ make FTH the default extension
                BL WORD COUNT DEFEXTMAX MIN DEFEXT$ PLACE ;

CREATE OPENBUF  \ a place to manipulate the filename
                MAXSTRING ALLOT

: "TO-PATHEND"  ( a1 n1 --- a2 n2 )     \ return a2 and count=n1 of filename
                OVER 1+ C@ [CHAR] : =   \ second char is ':'
                OVER 2 > AND            \ and name is longer than two characters
                IF      2 /STRING       \ then remove first two characters
                THEN                    \ now scan to end of last '\' in filename
                BEGIN   2DUP [CHAR] \ SCAN ?DUP
                WHILE   2SWAP 2DROP 1 /STRING
                REPEAT  DROP ;

: ?DEFEXT       ( addr -- )       \ conditionally add a default extension
                EXT_ADDED? OFF
                DEFEXT_ON? @
                IF      DUP COUNT "TO-PATHEND"  \ strip off path from filename
                        [CHAR] . SCAN NIP 0=    \ skip to file extension after '.',
                                                \ if not file extension was specified
                        DEFEXT$ C@ 0<> AND      \ and DEFEXT$ is not NULL 
                        IF      >R
                                S" ."         R@ +PLACE \ then lay down a decimal point
                                DEFEXT$ COUNT R@ +PLACE \ and lay in default file extension
                                R>
                                EXT_ADDED? ON
                        THEN
                THEN    DROP ;

: _"OPEN        ( a1 n1 -- fileid f1 )  \ open filename a1,n1
                                        \ return fileid and f1=false=ok
                OPENBUF PLACE                           \ drop name to OPENBUF
                OPENBUF ?DEFEXT                         \ add extension if needed
                OPENBUF COUNT 0 FOPEN-FILE              \ try to open it
                DUP 0=                                  \ if we succeeded
                IF      OPENBUF COUNT CUR-FILE PLACE    \ then set current file
                THEN    OPENBUF COUNT POCKET PLACE ;    \ and set POCKET

DEFER "OPEN     ( a1 n1 -- fileid f1 )  \ open filename a1,n1
   ' _"OPEN IS "OPEN                    \ return fileid and f1=false=ok

: $OPEN         ( addr -- fileid f1 )   \ open counted filename specified by addr
                                        \ return fileid and f1=false=ok
                COUNT "OPEN ;

DEFER START-INCLUDE     ' NOOP IS START-INCLUDE
DEFER   END-INCLUDE     ' NOOP IS   END-INCLUDE

: INCLUDE-FILE  ( fileid -- ) \ load file open on "fileid" to current dictionary
                MAXSTRING _LOCALALLOC >R        \ a place to save TIB contents
                TIB R@ MAXSTRING MOVE           \ save input line
                LOADFILE @ CELL+ >R             \ save pointer to filename
                ?LOADING @ >R
                LOADLINE @ >R
                >IN @ >R
                SOURCE-ID >R
                ( fileid ) TO SOURCE-ID
                SOURCE 2>R                      \ save current source
                SOURCE-POSITION >R
                ?LOADING ON                     \ mark as loading a file
                POCKET LINKFILE                 \ create a filename link
                LOADLINE OFF                    \ clear the loadline counter
                0 TO SOURCE-POSITION            \ reset the loadfile position
                START-INCLUDE
                ['] DO-INCLUDE CATCH            \ load file and catch errors
                SOURCE-ID FCLOSE-FILE DROP
                R> TO SOURCE-POSITION
                2R> (SOURCE) 2!                 \ restore input source buffer
                R> TO SOURCE-ID
                END-INCLUDE

                THROW                           \ throw load error if any
                R> >IN !
                R> LOADLINE !
                R> ?LOADING !
                ALIGN
                R> LINKFILE
                R> TIB MAXSTRING MOVE           \ recover last input line
                ?LOADING @
                IF      LOADFILE @ CELL+        \ step to most recent file
                        COUNT "CLIP"            \ get addr and length
                        CUR-FILE PLACE          \ make current again
                THEN     _LOCALFREE ;

: INCLUDED      ( addr len -- )   \ load file addr,len into current dictionary
                "OPEN ABORT" file not found"  INCLUDE-FILE ;

: FLOAD         ( -<filename>- )  \ load "filename" into application dictionary
                BL WORD COUNT INCLUDED ;

: SYS-FLOAD     ( -<filename>- )  \ load "filename" into system dictionary
                IN-SYSTEM? >R IN-SYSTEM
                BL WORD COUNT INCLUDED
                R> 0= IF IN-APPLICATION THEN ;

: INCLUDE       ( -<filename>- )  \ load "filename" into application dictionary
                FLOAD ;

: OK            ( -- )
                CUR-FILE COUNT INCLUDED ;

: ((            ( -- )
                BEGIN   BL WORD DUP @
                        [ HEX FFFFFF DECIMAL ] LITERAL AND
                        [ HEX 292902 DECIMAL ] LITERAL <>
                WHILE   C@ 0=
                        IF      REFILL
                                0= ABORT" missing ))"
                        THEN
                REPEAT  DROP ; IMMEDIATE

: \S            \ comment to end of file
                FALSE TO INCLUDING?  [COMPILE] \ ; IMMEDIATE

\ -------------------- Text Interpreter Loop --------------------

\ ( -- )  start editor at error
DEFER EDIT-ERROR   ' NOOP IS EDIT-ERROR

: _RESET-STACKS ( ?? -- )
                SP0 @ SP! ;

DEFER RESET-STACKS   ' _RESET-STACKS IS RESET-STACKS

: _MESSAGE      ( n -- )
                BASE @ >R DECIMAL
                CR ." Error: "
                POCKET COUNT TYPE SPACE
                DUP THROW_ABORTQ =
                IF      DROP MSG @ COUNT TYPE
                ELSE    ." Error # " .
                THEN
                ?LOADING @
                IF      BASE @ >R DECIMAL
                        CR ." File: "
                        LOADFILE @ CELL+ COUNT TYPE
                        ."  at line: "
                        LOADLINE ?
                        R> BASE !
                        EDIT-ERROR
                THEN
                R> BASE ! ;

DEFER MESSAGE   ' _MESSAGE IS MESSAGE

: QUERY-INTERPRET   ( -- )
                QUERY SPACE INTERPRET ;

: _QUIT         ( -- )
                RP0 @ RP!
                BEGIN   [COMPILE] [
                        ?LOADING OFF

                        BEGIN   CR ['] QUERY-INTERPRET CATCH  ?DUP 0=
                        WHILE   STATE @ 0=
                                IF      ."  ok"
                                        DEPTH .SMAX @ MIN 0
                                        ?DO  [CHAR] . EMIT  LOOP
                                THEN
                        REPEAT
                        CONSOLE         \ select the forth console
                        DUP 1+          \ no message on abort
                        IF      MESSAGE
                        THEN
                        RESET-STACKS    \ reset the stacks
                AGAIN ;

: DOCMDTAIL     ( -- )
                CMDLEN 2@ EVALUATE ;

: /IMAGE        ( 'imagename' -- )      \ dummy, really loaded by wrapper
                BL WORD DROP ;          \ just discard the name following

DEFER QUIT              ' _QUIT         IS QUIT
DEFER COMMANDLINE       ' DOCMDTAIL     IS COMMANDLINE
DEFER BOOT              ' INIT-SCREEN   IS BOOT

: COLD          ( -- )
                UNLOADFORTH REL>ABS &UNLOAD-FORTH !
                ['] BOOT CATCH
                IF      BYE         ( fatal error, exit)
                THEN
                &EXCEPT @ 0=
                IF      ['] COMMANDLINE CATCH DUP
                        IF      DUP 1+ IF MESSAGE THEN
                        THEN
                THEN
                RESET-STACKS
                SOURCE-ID
                IF      SOURCE-ID -1 <>
                        IF      SOURCE-ID FCLOSE-FILE DROP
                        THEN    0 TO SOURCE-ID
                                0 TO SOURCE-POSITION
                THEN
                QUIT ;

: DLLCOLD       ( -- )
                UNLOADFORTH REL>ABS &UNLOAD-FORTH !
                BOOT ;

\ -------------------- Class Primatives --------------------

CODE SKIP       ( adr len char -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repz    scasb
                je      short @@1
                inc     ecx
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

CODE SCAN       ( adr len char -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repnz   scasb
                jne     short @@1
                inc     ecx
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

CODE WSKIP      ( adr len word -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repz    scasw
                je      short @@1
                inc     ecx
                dec     edi
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

CODE WSCAN      ( adr len word -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repnz   scasw
                jne     short @@1
                inc     ecx
                dec     edi
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

CODE LSKIP      ( adr len long -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repz    scasd
                je      short @@1
                inc     ecx
                dec     edi
                dec     edi
                dec     edi
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

CODE LSCAN      ( adr len long -- adr' len' )
                pop     ecx
                jecxz   short @@2
                mov     eax, ebx                \ eax = character
                mov     ebx, edi                \ ebx = forth base address
                pop     edi
                add     edi, ebx                \ edi = absolute address
                repnz   scasd
                jne     short @@1
                inc     ecx
                dec     edi
                dec     edi
                dec     edi
                dec     edi
@@1:            sub     edi, ebx                \ relative
                push    edi
                mov     edi, ebx                \ restore base ptr
@@2:            mov     ebx, ecx
                next    c;

\ Scan for char BACKWARDS starting at addr, back through len bytes
\ before addr, returning addr' and len' of char.

CODE -SCAN      ( addr len char -- addr' len' )
                mov     eax, ebx
                pop     ecx
                jecxz   short @@1
                mov     ebx, edi
                pop     edi
                add     edi, ebx                \ edi = absolute address
                std
                repnz   scasb
                cld
                jne     short @@2
                inc     ecx
                inc     edi
@@2:            sub     edi, ebx
                push    edi
                mov     edi, ebx
@@1:            mov     ebx, ecx
                next    c;

\ Skip occurances of char BACKWARDS starting at addr, back through
\ addr-len, returning addr' and len' of char.

CODE -SKIP      ( addr len char -- addr' len' )
                mov     eax, ebx
                pop     ecx
                jecxz   short @@1
                mov     ebx, edi
                pop     edi
                add     edi, ebx                \ edi = absolute address
                std
                repz    scasb
                cld
                je      short @@2
                inc     ecx
                inc     edi
@@2:            sub edi, ebx
                push    edi
                mov     edi, ebx
@@1:            mov     ebx, ecx
                next    c;

\ COMPARE compares two strings. The return value is:
\
\        0 = string1 = string2
\       -1 = string1 < string2
\        1 = string1 > string2

CODE COMPARE    ( adr1 len1 adr2 len2 -- n )
                sub     ebp, # 8
                mov     0 [ebp], edi
                mov     4 [ebp], esi
                pop     eax                     \ eax = adr2
                pop     ecx                     \ ecx = len1
                pop     esi                     \ esi = adr1
                add     esi, edi                \ absolute address
                add     edi, eax                \ edi = adr2 (abs)
                sub     eax, eax                \ default is 0 (strings match)
                cmp     ecx, ebx                \ compare lengths
                je      short @@2
                ja      short @@1
                dec     eax                     \ if len1 < len2, default is -1
                jmp     short @@2
@@1:            inc     eax                     \ if len1 > len2, default is 1
                mov     ecx, ebx                \ and use shorter length
@@2:            repz cmpsb                      \ compare the strings
                je      short @@4               \ if equal, return default
                jns     short @@3
                mov     eax, # -1               \ if str1 < str2, return -1
                jmp     short @@4
@@3:            mov     eax, # 1                \ if str1 > str2, return 1
@@4:            mov     ebx, eax
                mov     edi, 0 [ebp]
                mov     esi, 4 [ebp]
                add     ebp, # 8
                next    c;

\ Search str1 for substring str2.
\ If found, return the address of the start of the
\ string, the characters remaining in str1 and a true flag.
\ Otherwise return the original str1 and a false flag.

\ ESI = pointer to source string (str2)
\ EBX = length  of source string
\ EDI = pointer to destination string (str1)
\ ECX = length  of destination string
\ EDX = pointer for compare

CODE SEARCH     ( adr1 len1 adr2 len2 -- adr3 len3 flag )
                test    ebx, ebx
                jne     short @@1
                pop     eax
                dec     ebx             \ zero length matches
                jmp     short @@6
@@1:            sub     ebp, # 12
                mov     0 [ebp], edx    \ save UP
                mov     4 [ebp], edi
                mov     8 [ebp], esi
                pop     esi
                add     esi, edi
                mov     ecx, 0 [esp]
                add     edi, 4 [esp]
                jmp     short @@2
@@4:            inc     edi             \ go to next    c; char in destination
                dec     ecx
@@2:            cmp     ecx, ebx        \ enough room for match?
                jb      short @@5
                sub     edx, edx        \ starting index
@@3:            mov     al, 0 [edx] [esi]
                cmp     al, 0 [edx] [edi]
                jne     short @@4
                inc     edx
                cmp     edx, ebx
                jne     short @@3
                mov     eax, edi        \ found
                mov     edx, 0 [ebp]
                mov     edi, 4 [ebp]
                mov     esi, 8 [ebp]
                add     ebp, # 12
                sub     eax, edi        \ relative address
                mov     4 [esp], eax
                mov     0 [esp], ecx
                mov     ebx, # -1       \ true flag
                jmp     short @@6
@@5:            sub     ebx, ebx        \ not found
                mov     edx, 0 [ebp]    \ restore UP
                mov     edi, 4 [ebp]
                mov     esi, 8 [ebp]
                add     ebp, # 12
@@6:            next    c;

CODE LARGEST    ( a1 n1 --- a2 n2 )
                mov     ecx, ebx          \ count of array to search
                pop     ebx               \ starting address of array
                push    edx               \ save UP
                xor     eax, eax          \ starting highest value = 0
                mov     edx, ebx          \ highest value addr = start address
@@1:            cmp     0 [ebx] [edi], eax  \ compare 32bit words
                jle     short @@2         \ branch if not greater
                mov     edx, ebx          \ if greater, save offset in EDX
                mov     eax, 0 [ebx] [edi]  \ and contents in EAX
@@2:            add     ebx, # 4          \ bump to next element
                dec     ecx
                jnz     @@1
                mov     ecx, edx
                pop     edx               \ restore UP
                push    ecx
                mov     ebx, eax
                next    c;

CODE CELL-SORT  ( a1 n1 -- )      \ perform in place sort buffer a1 of n1 cells
                push    ebx
                cmp     ebx, # 2        \ don't sort if less than 2 elements
                jnl     short @@2
                jmp     short @@3
@@1:            mov     eax,   0 [ebx] [edi]
                xchg    eax, 4 [ebx] [edi]
                xchg    eax,   0 [ebx] [edi]
                cmp     eax,   0 [ebx] [edi]
                jl      short @@1
                add     ebx, # 4
                loop    @@1
@@2:            pop     ecx
                pop     ebx
                push    ebx
                dec     ecx
                push    ecx
                jg      short @@1
@@3:            add     esp, # 8
                pop     ebx
                next    c;

CODE BYTE-SORT  ( a1 n1 -- )      \ perform in place sort buffer a1 of n1 bytes
                push    ebx
                cmp     ebx, # 2        \ don't sort if less than 2 elements
                jnl     short @@2
                jmp     short @@3
@@1:            mov     al,   0 [ebx] [edi]
                xchg    al, 1 [ebx] [edi]
                xchg    al,   0 [ebx] [edi]
                cmp     al,   0 [ebx] [edi]
                jl      short @@1
                inc     ebx
                loop    @@1
@@2:            pop     ecx
                pop     ebx
                push    ebx
                dec     ecx
                push    ecx
                jg      short @@1
@@3:            add     esp, # 8
                pop     ebx
                next    c;

\ add two numbers and return true if they would produce arithemtic
\ overflow (used by the debugger).

CODE +OV?       ( n1 n2 -- f )
                sub     ecx, ecx
                pop     eax
                add     ebx, eax
                jno     short @@1
                dec     ecx
@@1:            mov     ebx, ecx
                next    c;

\ -------------------- Local Variable Runtime --------------------

LABEL LOCAL@
                mov     ecx, LP [UP]  
                mov     eax, 4 [eax] [edi]
                push    ebx
                mov     ebx, 0 [eax] [ecx]
                next    c;

LABEL LOCAL!
                mov     ecx, LP [UP]  
                mov     eax, -4 [eax] [edi]
                mov     0 [eax] [ecx], ebx
                pop     ebx
                next    c;

LABEL LOCAL+!
                mov     ecx, LP [UP]    
                mov     eax, -8 [eax] [edi]
                add     0 [eax] [ecx], ebx
                pop     ebx
                next    c;

ASSEMBLER LOCAL@  META CONSTANT DOLOCAL
ASSEMBLER LOCAL!  META CONSTANT DOLOCAL!
ASSEMBLER LOCAL+! META CONSTANT DOLOCAL+!

 0 LOCAL LOCAL0
 1 LOCAL LOCAL1
 2 LOCAL LOCAL2
 3 LOCAL LOCAL3
 4 LOCAL LOCAL4
 5 LOCAL LOCAL5
 6 LOCAL LOCAL6
 7 LOCAL LOCAL7
 8 LOCAL LOCAL8
 9 LOCAL LOCAL9
10 LOCAL LOCAL10
11 LOCAL LOCAL11

\ allocate locals/args

CREATE DOCOLP
                INIT-ASSEMBLER
                xchg    esp, ebp
                push    esi
                push    LP [UP]
                xchg    esp, ebp
                mov     LP [UP]  , ebp
                lea     esi, 4 [eax] [edi]
                sub     eax, eax
                lodsb                           \ get # locals
                lea     eax, 0 [eax*4]
                sub     ebp, eax                \ allocate (not initialized)
                lodsb                           \ get # args
                and     eax, # 0xFF
                jz      short @@2
@@1:            sub     ebp, # 4
                mov     0 [ebp], ebx
                pop     ebx
                dec     eax
                jnz     @@1
@@2:            add     esi, # 2                \ align
                next    c;

NCODE EXITP
                mov     ebp, LP [UP]  
                xchg    esp, ebp
                pop     LP [UP]    
                pop     esi
                xchg    esp, ebp
                next    c;

NCODE UNNESTP
                mov     ebp, LP [UP]  
                xchg    esp, ebp
                pop     LP [UP]    
                pop     esi
                xchg    esp, ebp
                next    c;

NCODE INIT-LOCALS ( loc1 loc2 ... -- )
                sub     ebp, # 4
                mov     eax, LP [UP]              \ get old local pointer
                mov     0 [ebp], eax            \ save olp local pointer
                mov     LP [UP]  , ebp            \ save new local pointer
                sub     eax, eax                \ clear high bytes
                lodsb                           \ get # of uninitialized locals
                lea     eax, 0 [eax*4]          \ multiply by 4
                sub     ebp, eax                \ allocate non-initialized
                lodsb                           \ get # of initialized locals
                and     eax, # 0xFF
                jz      short @@2
@@1:            sub     ebp, # 4
                mov     0 [ebp], ebx            \ push on return stack
                pop     ebx                     \ get next    c; parameter
                dec     eax
                jnz     @@1
@@2:            add     esi, # 2                \ align
                next    c;

12 CONSTANT #-LOCALS

CREATE LOCAL-PTRS ' LOCAL0 , ' LOCAL1 , ' LOCAL2 , ' LOCAL3 , ' LOCAL4 ,
                  ' LOCAL5 , ' LOCAL6 , ' LOCAL7 , ' LOCAL8 , ' LOCAL9 ,
                  ' LOCAL10 , ' LOCAL11 ,

: >LOC  ( n -- cfa )  CELLS LOCAL-PTRS + @ ;

\ -------------------- Parameter Name List --------------------

\ Names in Win32For are limited in length to name-max-chars characters.
\ Parameter names are stored as name-max-chars byte counted string.
\ PARMLIST holds #-LOCALS plus one names of locals.  The first slot is
\ used as a temp for searching and re-ordering local variable names.

CREATE PARMLIST
        #-LOCALS 1+ NAME-MAX-CHARS 1+ * ALLOT   \ list of parameter names
                                \ first slot is temp for current name

: "FIRSTPARM    ( addr cnt -- ) \ store string in first slot of parmlist
                NAME-MAX-CHARS MIN PARMLIST PLACE ;

: ADDPARM       ( addr cnt -- ) \ add string to end of parmlist
                PARMS #-LOCALS 1- > ABORT" Too many parameters"
                1 +TO PARMS
                2DUP "FIRSTPARM
                NAME-MAX-CHARS MIN PARMLIST PARMS
                NAME-MAX-CHARS 1+ * + PLACE ;

\ -------------------- Parameter Compiler --------------------

0 VALUE INPARMS         \ number of input paramters
0 VALUE LOCFLG          \ 1 = compiling args, 0 = compiling locals

: PARMS,        ( -- )   \ compile runtime to push parameters
        ?:M                             \ in method?
        IF      CELL NEGATE ALLOT       \ then deallocate the cell layed down
        ELSE    COMPILE INIT-LOCALS     \ else this is a normal local def
        THEN
        INPARMS PARMS OVER - C, ( #locals ) C, ( #args )
        0 C, 0 C,  ( unused filler bytes for cell alignment )   ;

: FIRSTCHR      ( addr cnt -- addr cnt chr )
                OVER C@ ;

: (LOCAL)       ( addr cnt -- )
                DUP 0=
                IF      2DROP
                ELSE    ADDPARM
                        INPARMS LOCFLG + TO INPARMS
                THEN    ;

\ August 2nd, 1999 - 11:13 tjz
\ modfied versin of a word suggested by Robert Smith, to get a word from the
\ input stream, delimited by 'char', even if a line crossing is needed.

: NEXTWORD      ( char -- adr flag ) \ flag=TRUE if we got a word, else FALSE
                BEGIN   DUP WORD DUP C@ 0=
                WHILE   refill 0=
                        IF      NIP     \ discard 'char'
                                FALSE   \ can't find a word
                                _EXIT   \ time to leave
                        ELSE    DROP    \ discard empty word
                        THEN
                REPEAT  NIP TRUE ;

\ August 2nd, 1999 - 10:31 tjz
\ The following source lines were originally supposed to prevent the
\ creation of an extra local variable, when a 'name' type comment was
\ inserted into a local variable declaration. I now believe this was a
\ bad idea, and could prevent real local variables from being defined,
\ when they start with a ''' character.
\                FIRSTCHR DUP [CHAR] - <>       \ as in { name -- }
\                        SWAP [CHAR] ' <> AND   \ or as { name 'name' -- }

\ Also reworked, to allow local variables to be broken into multiple
\ lines, as suggested by Robert Smith.
\ February 11th, 2002 - 17:45  Changes made as per Tom Zimmer request.
\ This now handles the case of   : foo { a b c } ;

: {     ( -- )  \ begin local variable usage in the form;
                \ { initedloc1 initedloc2 \ uninitedloc3 -- comments }
        ?COMP
        PARMS ABORT" Locals may be defined only once per definition."
        0 TO INPARMS
        1 TO LOCFLG
        BEGIN   BL NEXTWORD 0=
                abort" EOF encountered while searching for '}' "
                ?UPPERCASE COUNT
                FIRSTCHR [CHAR] - = >R          \ as in { name -- }

                FIRSTCHR [CHAR] } = R> OR 0=    \ or in { name }
                \ IF neither, then not done
        WHILE   FIRSTCHR [CHAR] \ =             \ start of non-initialized
                                                \ local variables as in
                                                \ { name \ another -- }
                IF      0 TO LOCFLG 2DROP
                ELSE    FIRSTCHR [CHAR] } = ABORT" Args missing --"

                        (LOCAL)
                THEN
        REPEAT  DROP C@ [CHAR] } <> >R
        PARMS 0>
        IF      PARMS,  ( compile runtime code )
        THEN    R>
        IF      BEGIN   BL NEXTWORD 0= ABORT" Args missing }"
                        COUNT FIRSTCHR [CHAR] } = NIP NIP
                UNTIL
        THEN    ; IMMEDIATE

: REVERSEARGS   ( -- )
                INPARMS ?DUP
                IF      2/ 0
                        ?DO     PARMLIST I 1+ NAME-MAX-CHARS 1+ * +
                                PARMLIST NAME-MAX-CHARS 1+ MOVE
                                PARMLIST INPARMS I - NAME-MAX-CHARS 1+ * +
                                PARMLIST I 1+ NAME-MAX-CHARS 1+ * +
                                NAME-MAX-CHARS 1+ MOVE
                                PARMLIST
                                PARMLIST INPARMS I - NAME-MAX-CHARS 1+ * +
                                NAME-MAX-CHARS 1+ MOVE
                        LOOP
                THEN    ;

: LOCALS|       ( -- )
                ?COMP
                PARMS ABORT" Locals may be defined only once per definition."
                0 TO PARMS
                0 TO INPARMS
                1 TO LOCFLG
                BEGIN
                        BL WORD COUNT 2DUP UPPER
                        FIRSTCHR [CHAR] | <>
                WHILE
                        (LOCAL)
                REPEAT  2DROP
                PARMS 0>
                IF      PARMS,  ( compile runtime code )
                THEN
                REVERSEARGS ; IMMEDIATE

: &LOCAL        ( -<name>- a1 ) \ return the address of local "name"
                R> DUP CELL+ >R
                ABS>REL @ DUP @ DOLOCAL <>
                ABORT" Must be followed by a local variable"
                CELL+ @ LP @ ABS>REL + ;

: PFIND         ( addr -- addr FALSE | cfa -1 | cfa 1 )
                FALSE                   \ flag initially not found
                STATE @ 0=              \ if not compiling
                PARMS 1 < OR            \ or no parameters
                IF      _EXIT           \ then don't try to find local
                THEN
                OVER COUNT "FIRSTPARM   \ put in first parm slot
                PARMS 1+ 1
                DO      PARMLIST I NAME-MAX-CHARS 1+ * + COUNT
                        PARMLIST COUNT
                        COMPARE 0=                      \ TRUE if matched
                        IF      2DROP  PARMS  I  -  >LOC TRUE
                                LEAVE
                        THEN
                LOOP    ;

: PARMFIND      ( addr -- addr FALSE | cfa -1 | cfa 1 )
                DUP COUNT FIND-BUFFER PLACE
                FIND-BUFFER ?UPPERCASE
                PFIND ?DUP 0= IF (FIND)  THEN
                DUP>R
                IF      NIP
                ELSE    DROP
                THEN     R> ;

' PARMFIND IS FIND

\ ---------------- 32-bit Hash Function for Objects ----------------

CODE METHOD-HASH ( addr len -- hash-val )
                mov     ecx, ebx                \ ecx = count
                pop     ebx                     \ ebx = address
                mov     eax, ecx                \ eax = result (includes count)
                jecxz   short @@3
@@1:            rol     eax, # 7
                xor     al, 0 [ebx] [edi]       \ add in next    c; byte
                inc     ebx
                dec     ecx
                jnz     @@1
@@3:            test    eax, eax                \ make sure it's negative
                js      short @@2
                not     eax
@@2:            mov     ebx, eax
                next    c;

\ -------------------- Method/Ivar Search --------------------

\ : ((FINDM))   ( SelID addr -- 0cfa t | f )
\         Begin   @  dup
\         while   2dup cell+ @ =
\                 if      nip  2 cells+  true _exit
\                 then
\         repeat  nip ;

CODE ((FINDM))  ( SelID addr -- 0cfa t | f )
                pop     eax
@@1:            mov     ebx, 0 [ebx] [edi]      \ follow link
                test    ebx, ebx                \ end of list?
                je      short @@2
                cmp     eax, 4 [ebx] [edi]      \ selectors match?
                jne     short @@1
                add     ebx, # 8                \ method cfa
                push    ebx
                mov     ebx, # -1               \ and true flag
@@2:            next    c;

\ -------------------- Runtime for Methods --------------------
\ m0cfa is executed when the object address is on the stack
\ m1cfa is executed when the ivar offset is compile in-line

CREATE M0CFA
        0 OFA-H !               \ disable OFA resolution
                INIT-ASSEMBLER
                xchg    esp, ebp
                push    esi
                push    OP [UP]  
                push    LP [UP]    
                xchg    esp, ebp
                mov     LP [UP]  , ebp          \ load new local pointer
                lea     esi, 8 [eax] [edi]      \ get new ip (skip m1cfa)
                mov     OP [UP]  , ebx            \ get object address
                pop     ebx
                sub     eax, eax
                lodsb                           \ get # locals
                lea     eax, 0 [eax*4]
                sub     ebp, eax                \ allocate (not initialized)
                lodsb                           \ get # args
                and     eax, # 0xFF
                jz      short @@2
@@1:            sub     ebp, # 4
                mov     0 [ebp], ebx
                pop     ebx
                dec     eax
                jnz     @@1
@@2:            add     esi, # 2                \ align
                next    c;

CREATE M1CFA
        0 OFA-H !               \ disable OFA resolution
                INIT-ASSEMBLER
                mov     ecx, 0 [esi]            \ get inline offset
                add     esi, # 4
                xchg    esp, ebp
                push    esi
                push    OP [UP]
                push    LP [UP]  
                xchg    esp, ebp
                mov     LP [UP]  , ebp            \ load new local pointer
                lea     esi, 4 [eax] [edi]      \ get new ip
                add     OP [UP]  , ecx            \ add offset to object address
                sub     eax, eax
                lodsb                           \ get # locals
                lea     eax, 0 [eax*4]
                sub     ebp, eax                \ allocate (not initialized)
                lodsb                           \ get # args
                and     eax, # 0xFF
                jz      short @@2
@@1:            sub     ebp, # 4
                mov     0 [ebp], ebx
                pop     ebx
                dec     eax
                jnz     @@1
@@2:            add     esi, # 2                \ align
                next    c;

\ end of method definition

NCODE EXITM     ( -- )
                mov     ebp, LP [UP]  
                xchg    esp, ebp
                pop     LP [UP]                   \ restore local pointer
                pop     OP [UP]                   \ restore object pointer
                pop     esi                     \ restore ip
                xchg    esp, ebp
                next    c;

NCODE UNNESTM   ( -- )
                mov     ebp, LP [UP]  
                xchg    esp, ebp
                pop     LP [UP]                   \ restore local pointer
                pop     OP [UP]                   \ restore object pointer
                pop     esi                     \ restore ip
                xchg    esp, ebp
                next    c;

NCODE _LOCALFREE ( -- )                         \ release local allocation
                mov     ebp, LP [UP]  
                xchg    esp, ebp
                pop     LP [UP]  
                xchg    esp, ebp
                next    c;

NCODE _LOCALALLOCP ( n1 -- a1 )
                sub     ebp, ebx        \ subtract n1 from return stack
                and     ebp, # -4       \ cell align return stack
                mov     ebx, ebp        \ move to top of stack
                sub     ebx, edi        \ convert from absolute to relative
                next    c;

NCODE _LOCALALLOC ( n1 -- a1 )
                xchg    esp, ebp
                push    LP [UP]  
                xchg    esp, ebp
                mov     LP [UP]  , ebp

                sub     ebp, ebx        \ subtract n1 from return stack
                and     ebp, # -4       \ cell align return stack
                mov     ebx, ebp        \ move to top of stack
                sub     ebx, edi        \ convert from absolute to relative
                next    c;

: LOCALALLOC    ( n1 -- a1 )            \ allocate n1 bytes of return stack
                                        \ return a1 the address of the array
                PARMS                   \ if no locals, setup stack frame
                IF      COMPILE _LOCALALLOCP
                ELSE    -1 TO PARMS
                        COMPILE _LOCALALLOC
                THEN    ; IMMEDIATE

\ -------------------- Runtime for Objects --------------------

CREATE DOOBJ    ( -- addr )
                INIT-ASSEMBLER
                push    ebx
                lea     ebx, 8 [eax]
                next    c;

\ return the base of the current object

CODE ^BASE      ( -- addr )
                push    ebx
                mov     ebx, OP [UP]    
                next    c;

\ -------------------- Runtime for Instance Variables --------------------

\ These are the "non-object" instance variables.  They have the same
\ names and syntax as the regular Forth versions.  They are defined in
\ the META vocabulary and will be found when inside a class definition.
\ The Offset from OP stored at cfa+15

\ bytes runtime

CREATE (&IV)
                INIT-ASSEMBLER
                push    ebx
                mov     ebx, 4 [eax] [edi]      \ get ivar offset
                add     ebx, OP [UP]            \ add to base of current object
                next    c;

\ single byte (8bit) instance variables

CREATE (IVC@)
                INIT-ASSEMBLER
                push    ebx
                mov     eax, 4 [eax] [edi]      \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     bl,  0 [eax] [edi]      \ fetch value
                and     ebx, # 0xFF
                next    c;

CREATE (IVC!)
                INIT-ASSEMBLER
                mov     eax, -4 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     0 [eax] [edi], bl       \ store value
                pop     ebx
                next    c;

CREATE (IVC+!)
                INIT-ASSEMBLER
                mov     eax, -8 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                add     0 [eax] [edi], bl       \ add value
                pop     ebx
                next    c;

\ word number (16bit) instance variables

CREATE (IVW@)
                INIT-ASSEMBLER
                push    ebx
                mov     eax, 4 [eax] [edi]      \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     bx, 0 [eax] [edi]       \ fetch WORD value
                and     ebx, # 0xFFFF
                next    c;

CREATE (IVW!)
                INIT-ASSEMBLER
                mov     eax, -4 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     0 [eax] [edi], bx       \ store WORD value
                pop     ebx
                next    c;

CREATE (IVW+!)
                INIT-ASSEMBLER
                mov     eax, -8 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                add     0 [eax] [edi], bx       \ add WORD value
                pop     ebx
                next    c;

\ single number (32bit) instance variables

CREATE (IV@)
                INIT-ASSEMBLER
                push    ebx
                mov     eax, 4 [eax] [edi]      \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     ebx, 0 [eax] [edi]      \ fetch value
                next    c;

CREATE (IV!)
                INIT-ASSEMBLER
                mov     eax, -4 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                mov     0 [eax] [edi], ebx      \ store value
                pop     ebx
                next    c;

CREATE (IV+!)
                INIT-ASSEMBLER
                mov     eax, -8 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]            \ add to base of current object
                add     0 [eax] [edi], ebx      \ add value
                pop     ebx
                next    c;

\ double number (64bit) instance variable

CREATE (IVD@)
                INIT-ASSEMBLER
                push    ebx
                mov     eax, 4 [eax] [edi]      \ get ivar offset
                add     eax, OP [UP]              \ add to base of current object
                mov     ecx, 4 [eax] [edi]      \ fetch value at cell +
                mov     ebx, 0 [eax] [edi]      \ fetch value
                push    ecx
                next    c;

CREATE (IVD!)
                INIT-ASSEMBLER
                mov     eax, -4 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]              \ add to base of current object
                mov     0 [eax] [edi], ebx      \ store value
                pop     ebx
                mov     4 [eax] [edi], ebx      \ store value
                pop     ebx
                next    c;

CREATE (IVD+!)
                INIT-ASSEMBLER
                mov     eax, -8 [eax] [edi]     \ get ivar offset
                add     eax, OP [UP]              \ add to base of current object
                pop     ecx
                add     4 [eax] [edi], ecx      \ add low part first
                adc     0 [eax] [edi], ebx      \ add high part with carry
                pop     ebx
                next    c;

\ string instance variable

\ CREATE (IV"@)           \ fetch strings current contents
\         INIT-ASSEMBLER
\         push    ebx
\         mov     eax, 4 [eax] [edi]      \ get ivar offset
\         add     eax, OP [UP]              \ add to base of current object
\         sub     ebx, ebx
\         mov     bl, 0 [eax] [edi]       \ fetch the count
\         inc     eax                     \ bump to addr of first char
\         push    eax                     \ put back on stack
\         next    c;

\ CREATE (IV"!)           \ store string into instance variable
\         INIT-ASSEMBLER
\         push    ebx
\         mov     ebx, -4 [eax] [edi]     \ get ivar offset
\         add     ebx, OP [UP]              \ add to base of current object
\         lea     eax, ' PLACE            \ get address of PLACE
\         exec    c;                      \ go execute it

\ CREATE (IV"+!)          \ append string to current contents
\         INIT-ASSEMBLER
\         push    ebx
\         mov     ebx, -8 [eax] [edi]     \ get ivar offset
\         add     ebx, OP [UP]              \ add to base of current object
\         lea     eax, ' +PLACE           \ get address of +PLACE
\         exec    c;                      \ go execute it

\ Arrays have the index on the stack at runtime

\ CREATE (IV[]@)
\         INIT-ASSEMBLER
\         mov     eax, 4 [eax] [edi]      \ get ivar offset
\         add     eax, OP [UP]              \ add to base of current object
\         add     eax, edi
\         mov     ebx, 0 [eax] [ebx*4]
\         next    c;

\ CREATE (IV[]!)
\         INIT-ASSEMBLER
\         pop     ecx
\         mov     eax, -4 [eax] [edi]     \ get ivar offset
\         add     eax, OP [UP]              \ add to base of current object
\         add     eax, edi
\         mov     0 [eax] [ebx*4] , ecx
\         pop     ebx
\         next    c;

\ CREATE (IV[]+!)
\         INIT-ASSEMBLER
\         pop     ecx
\         mov     eax, -4 [eax] [edi]     \ get ivar offset
\         add     eax, OP [UP]              \ add to base of current object
\         add     eax, edi
\         add     0 [eax] [ebx*4] , ecx
\         pop     ebx
\         next    c;

\ --------------------------------------------------------------------------

DEFER SUBCLASS-WNDPROC  ' NOOP IS SUBCLASS-WNDPROC

CREATE SUBCLASS-RETURN
                HERE CELL+ ,-T
                HERE CELL+ ,-T
                INIT-ASSEMBLER
                mov     eax, ebx                \ C return value
                mov     esp, ebp                \ restore stack
                pop     ebp                     \ restore registers
                pop     ebx
                pop     edi
                pop     esi
                ret     # 4 CELLS               \ return & discard params
END-CODE

CREATE SUBCLASS-ENTRY  ( lparam wparam message hwnd -- result )
                INIT-ASSEMBLER
                pop     eax                     \ thunk address+8
                push    esi                     \ save registers
                push    edi
                push    ebx
                push    ebp
                mov     ebp, esp                \ make forth stacks
                sub     esp, # 4000             \ room for return stack
                push    5 CELLS [ebp]           \ hwnd
                push    6 CELLS [ebp]           \ message
                push    7 CELLS [ebp]           \ wparam
                push    8 CELLS [ebp]           \ lparam
                mov     edi, eax
                sub     edi, 0 [edi]            \ edi has forth base address
                lea     edx, CONUSER FOLLOWER RP0 - + [edi] \ init user pointer
                mov     ebx, eax                \ thunk+8
                sub     ebx, # 12               \ base of object
                sub     ebx, edi                \ make relative address
                lea     esi, SUBCLASS-RETURN [edi]
                mov     eax, # ' SUBCLASS-WNDPROC
                EXEC
END-CODE

\ --------------------------------------------------------------------------

\ WndProc allows us to create a window procedure that is a Forth word.
\
\ WndProc structure:
\
\       nop                     \ alignment
\       nop
\       nop
\       call    wndproc_begin
\       dd      here            \ to get relative address
\       dd      cfa             \ cfa of word to execute

LABEL WNDPROC_RETURN
                HERE CELL+ ,-T                  \ ITC thread
                HERE CELL+ ,-T
                mov     eax, ebx                \ return value
                mov     esp, ebp
                pop     SP0 [UP]    
                pop     ebp
                pop     ebx
                pop     edi
                pop     esi
                ret     # 4 CELLS
END-CODE

LABEL WNDPROC_BEGIN
                pop     eax                     \ thunk address
                push    esi
                push    edi
                push    ebx
                push    ebp
                mov     edi, eax
                mov     eax, 4 [edi]
                sub     edi, 0 [edi]
                lea     edx, CONUSER FOLLOWER RP0 - + [edi] \ init user pointer
                push    SP0 [UP]    
                mov     ebp, esp
                sub     esp, # 4000             \ room for return stack
                mov     SP0 [UP]  , esp
                sub     SP0 [UP]  , edi
                push         6 CELLS [ebp]      \ hwnd
                push         7 CELLS [ebp]      \ message
                push         8 CELLS [ebp]      \ wparam
                mov     ebx, 9 CELLS [ebp]      \ lparam
                lea     esi, WNDPROC_RETURN [edi]
                exec                            \ go for it!
END-CODE

: WNDPROC       ( cfa -- ) \ define window procedure
                CREATE
                [ ASSEMBLER WNDPROC_BEGIN META ] LITERAL CALL,
                HERE ,
                COMPILE, ;

LABEL UNLOAD-RETURN
                HERE CELL+ ,-T                  \ ITC thread
                HERE CELL+ ,-T
                mov     eax, ebx                \ return value
                mov     esp, ebp
                pop     ebp
                pop     ebx
                pop     edi
                pop     esi
                ret
END-CODE

LABEL DO-UNLOAD-FORTH
                pop     eax                     \ thunk address
                push    esi
                push    edi
                push    ebx
                push    ebp
                mov     ebp, esp
                sub     esp, # 4000             \ room for return stack
                mov     edi, eax                \ thunk address
                mov     eax, 4 [edi]            \ cfa of word to execute
                sub     edi, 0 [edi]              \ edi now has forth base address
                lea     edx, CONUSER FOLLOWER RP0 - + [edi] \ init user pointer
                lea     esi, UNLOAD-RETURN [edi]
                exec                            \ go for it!
END-CODE

CREATE UNLOADFORTH
                INIT-ASSEMBLER
                nop
                nop
                nop
                call do-unload-forth
                END-CODE
                HERE ,
                ' NOOP ,

\ -------------------- Tools --------------------

: .S            ( -- )
                ?STACK
                DEPTH .SMAX @ MIN DUP
                IF      ." ["
                        DEPTH 1- 1 .R
                        ." ] "
                        BEGIN   DUP PICK 1 .R
                                BASE @ 16 =
                                IF       ." h"
                                THEN
                                SPACE
                                1- DUP 0=
                        UNTIL
                ELSE    ."  empty "
                THEN    DROP ;

: NFA-COUNT     ( nfa -- adr len )
                DUP C@ 63 AND SWAP OVER - SWAP ;

: WAIT          ( -- )
                KEY ( k_ESC ) 27 = IF ABORT THEN ;

18 VALUE SCREENDELAY    \ delay value for some screen output

CREATE DELAYS     0 W,    1 W,   18 W,   25 W,   40 W,
                 60 W,   90 W,  120 W,  200 W,  500 W,

: _START/STOP   ( -- )
                KEY?
                IF KEY  10 DIGIT ( number keys select delay )
                        IF 2 * DELAYS + W@ TO SCREENDELAY
                        ELSE  ( k_ESC ) 27 = IF ABORT THEN  WAIT
                        THEN
                THEN ;

DEFER START/STOP ' _START/STOP IS START/STOP

: NUF?          ( -- f1 )
                ['] START/STOP CATCH ;

: H.R           ( n1 n2 -- )    \ display n1 as a hex number right
                                \ justified in a field of n2 characters
                BASE @ >R HEX >R
                0 <# #S #> R> OVER - SPACES TYPE
                R> BASE ! ;

: H.N           ( n1 n2 -- )    \ display n1 as a HEX number of n2 digits
                BASE @ >R HEX >R
                0 <# R> 0 ?DO # LOOP #> TYPE
                R> BASE ! ;

: H.2           ( n1 -- ) 2 H.N ;               \ two digit HEX number
: H.4           ( n1 -- ) 4 H.N ;               \ four digit HEX number
: H.8           ( n1 -- ) 8 H.N ;               \ eight digit HEX number

: .NAME         ( xt -- )       \ show name, if can't find name, show address
                DUP >NAME DUP NAME> ['] [UNKNOWN] =     \ if not found
                IF      DROP [CHAR] " EMIT ." 0x" 1 H.R [CHAR] " EMIT SPACE
                ELSE    .ID DROP
                THEN    ;

: KWORDS        ( -- )
                CR CONTEXT @
                DUP VOC#THREADS >R
                HERE 500 + R@ CELLS MOVE        \ copy vocabulary up
                BEGIN   HERE 500 + R@ LARGEST DUP
                WHILE   DUP L>NAME NFA-COUNT NIP 7 + ?CR
                        DUP LINK> ." 0x" H.4 SPACE
                        DUP L>NAME .ID 23 #TAB
                        @ SWAP !
                        START/STOP
                REPEAT  2DROP  R>DROP ;

: EMIT.         ( n -- )
                DUP BL 255 BETWEEN 0= IF  DROP [CHAR] .  THEN  EMIT ;

: +NO-WRAP      ( a1 n1 -- a2 ) \ add n1 to a1, limit to address 0xFFFFFFFF
                                \ don't allow wrap around to address zero
                0 TUCK D+ -1 0 DMIN DROP ;

DEFER DUMPC@    ' C@ IS DUMPC@

: DUMP          ( adr len -- )  ( hex byte format with ascii )
                OVER +NO-WRAP DUP ROT
                ?DO     CR I 4 U.R SPACE SPACE
                        I 16 +NO-WRAP OVER UMIN I
                        2dup
                        DO      I DUMPC@ H.2 space I J 7 + = IF SPACE THEN
                        LOOP    2DUP -  16 OVER - 3 *  SWAP 8 < -  1+ SPACES
                        DO      I DUMPC@ EMIT.
                        LOOP    NUF? ?LEAVE
                        SCREENDELAY MS          \ slow down output
            16 +LOOP    DROP    ;

: DM            ( a1 -- )
                -1 DUMP ;

\ =================================================================
\ =================================================================
\
\       These definitions must be last so they are not used
\       during meta compilation
\
\ =================================================================
\ =================================================================

: EXIT          ( -- )
                ?COMP
                ?:M     ( -- F1 )
                FALSE TO ?:M
                        ( -- f1 ) ABORT" Can't use EXIT in a Method !"
                PARMS
                IF      COMPILE EXITP
                ELSE    COMPILE _EXIT
                THEN    ; IMMEDIATE

: ?EXIT         ( F1 -- )
                ?COMP
                ?:M     ( -- F1 )
                FALSE TO ?:M
                        ( -- f1 ) ABORT" Can't use ?EXIT in a Method !"
                COMPILE ?BRANCH >MARK
                [COMPILE]  EXIT
                COMPILE _THEN
                >RESOLVE ; IMMEDIATE

: DOES>         ( -- )
                ?COMP
                ?:M     ( -- F1 )
                FALSE TO ?:M
                        ( -- f1 ) ABORT" Can't use DOES> in a Method !"
                PARMS
                IF      COMPILE _LOCALFREE
                        COMPILE PARMS-INIT
                THEN
                COMPILE (;CODE)
                DODOES CALL,
                PARMS-INIT ; IMMEDIATE

: #DOES>        ( -- )          \ "compile time"
                ( a1 -- )       \ "runtime" a1=cfa of word being defined
                ?COMP
                ?:M     ( -- F1 )
                FALSE TO ?:M
                        ( -- f1 ) ABORT" Can't use DOES> in a Method !"
                PARMS
                IF      COMPILE _LOCALFREE
                        COMPILE PARMS-INIT
                THEN
                COMPILE #(;CODE)
                DODOES CALL,
                PARMS-INIT ; IMMEDIATE

: LOCALALLOC:   ( n1 -<name>- )   \ allocate a local n1 byte buffer to local "name"
                ?COMP
                [COMPILE] LOCALALLOC [COMPILE] TO ; IMMEDIATE

\ --------------------------------------------------------------------------
\ --------------------------------------------------------------------------
\ Semicolon ';' must be compiled after its last use in the meta compile
\ --------------------------------------------------------------------------
\ --------------------------------------------------------------------------

DEFER DO-;CHAIN ' NOOP IS DO-;CHAIN

: ;             ( -- )
                ?COMP
                ?:M     ( -- F1 )
                FALSE TO ?:M
                        ( -- f1 ) ABORT" Methods must END in ;M !"
                ?CSP REVEAL
                PARMS
                IF      COMPILE UNNESTP
                ELSE    COMPILE UNNEST
                THEN    [COMPILE] [ 0 TO PARMS DO-;CHAIN ; IMMEDIATE

\ =================================================================
\ Initialization entry point, when Forth is being used as a DLL

LABEL calldllreturn
                HERE CELL+ ,-T          \ ITC thread
                HERE CELL+ ,-T
                mov     eax, ebx        \ return value
                mov     esp, ebp
                pop     ebp
                pop     ebx
                pop     edi
                pop     esi
                ret                     \ discard no bytes on return
END-CODE

LABEL calldllinit
                pop     eax                     \ thunk address
                push    esi
                push    edi
                push    ebx
                push    ebp
                mov     edi, eax                \ thunk address
                sub     edi, 0 [edi]              \ edi now has forth base address
                lea     edx, CONUSER FOLLOWER RP0 - + [edi] \ init user pointer
                mov     ebp, esp
                mov     eax, ebp
                sub     eax, edi
                mov     RP0 [UP]  , eax           \ save RP0
                mov     -4000 [esp], # 0        \ probe stack to force allocation
                mov     -8000 [esp], # 0
                sub     esp, # 8000             \ room for return stack

                mov     eax, esp
                sub     eax, edi
                mov     SP0 [UP]  , eax           \ save SP0
                pop     ebx                     \ top of stack in ebx

                mov     eax, # 10
                mov     BASE [UP]  , eax          \ default base to decimal

                mov     eax, # ' DLLCOLD        \ cfa of word to execute

                lea     esi, calldllreturn [edi]
                exec                            \ go for it!
END-CODE

LABEL DLLFORTHINIT              \ dll entry point for Forth initialization
HERE DLLINIT !-T
                nop
                nop
                nop
                call    CALLDLLINIT
END-CODE
                HERE ,

LABEL DLLFORTHENTRY                     \ dummy entry, patched later
HERE DLLNTRY !-T
                ret     # 4 CELLS       \ discard 4 arguments on return
END-CODE

\ -------------------- Cold Start Entry Point --------------------

LABEL _COLD1
                pop     edi
                sub     edi, 0 [edi]              \ edi = forth base
                lea     edx, CONUSER FOLLOWER RP0 - + [edi] \ init user pointer
                mov     ebp, esp                \ top of return stack
                mov     eax, ebp
                sub     eax, edi
                mov     RP0 [UP]  , eax           \ save RP0
                mov     -4000 [esp], # 0        \ probe stack to force allocation
                mov     -8000 [esp], # 0
                sub     esp, # 8000             \ room for return stack
                mov     eax, esp
                sub     eax, edi
                mov     SP0 [UP]  , eax           \ save SP0
                pop     ebx                     \ top of stack in ebx

                mov     eax, # 10
                mov     BASE [UP]  , eax          \ default base to decimal

                mov     eax, # ' COLD           \ execute cold
                exec
END-CODE

LABEL _COLD
HERE ENTRY !-T
        call    _COLD1
END-CODE
        HERE ,  ( relocation offset )

HERE-T ORIGIN -               AACTUAL !-T        \ set the actual dictionary length
HERE-H ORIGIN SSEPARATION + - SACTUAL !-T        \ set the actual header length

\ -------------------- The End --------------------

' THROW         RESOLVES THROW
' NFA-COUNT     RESOLVES NFA-COUNT
\ ' HERE          RESOLVES HERE 
' SCAN          RESOLVES SCAN
' UNLOADFORTH   RESOLVES UNLOADFORTH
' START/STOP    RESOLVES START/STOP
' -ALIGNED      RESOLVES -ALIGNED
' _LOCALALLOC   RESOLVES _LOCALALLOC
' _LOCALFREE    RESOLVES _LOCALFREE
' PARMFIND      RESOLVES PARMFIND
' .NAME         RESOLVES .NAME

\ Usage for ALIAS added March 11th, 2002 - 21:57 jp      
