\ $Id: library.f 1.1 1994/04/01 07:53:13 andrew Exp $

cr .( Loading Windows API Library...)

\ **********************************************************************
\
\ library format:
\       cell    handle
\       cell    linked list of libraries
\       cell    library type 0=normal, -1=system(no load for turnkey)
\       counted name (counted string)
\
\ proc format:
\       name
\       count
\       link to next proc
\       name (counted string)
\       procedure address
\       pointer to library
\       number of parameters
\
\ **********************************************************************

0 value ignore-missing-procs?

INTERNAL        \ non-userl definitions start here

variable winlib-link            \ linkage for libraries
         winlib-link off

variable winproc-link           \ linkage for procedures
         winproc-link off

in-system

: trim-procs    ( a1 -- a1 )    \ remove any procs above forget address
                dup winproc-link full-trim
                dup winlib-link  full-trim ;

\ forget unneeded procedures & libs
\in-system-ok forget-chain chain-add trim-procs

in-application

: do-link       { cfa-to-do link-address -- }
                begin   link-address @ dup to link-address
                while   link-address cfa-to-do execute
                repeat  ;

: load-winlibrary  ( lib -- f1 )        \ f1=FALSE if all is ok
        dup @ 0=                        \ library address NULL?
        if      dup 3 cells+            \ step from PFA to counted library name
                count swap rel>abs 16 xcall ( LoadLibrary )
                dup>r r@
                if      swap !          \ install library address
                else    2drop           \ discard and leave NULL
                then    r> 0=           \ -- f1
        else    drop    FALSE           \ if not, then return FALSE
        then    ;

: free-winlibrary  ( lib -- )
        dup @   17 xcall drop   off ;

EXTERNAL        \ switch to user accessable definitions

: "WinLibrary   { adr len -- }
        winlib-link @
        begin   ?dup
        while   dup 2 cells+ count adr len compare 0=
                if      drop EXIT       \ already in list, EXIT
                then
                @
        repeat
        in-system? abort" DLLs must be in APPLICATION dictionary
        0 ,                             \ the library handlehandle
        winlib-link link,               \ the link of all libraries
        0 ,                             \ mark as application library
        adr len here over 1+ allot place ;

: WinLibrary   { \ lib$ 'name.DLL' -- }  \ usage: WinLibrary user32<.dll>
        MAXSTRING LocalAlloc: lib$
        /parse-word count lib$ place
        lib$ count + 4 - 4 s" .DLL" caps-compare
        IF      s" .DLL" lib$ +place    \ if there is no .DLL file extension, add it
        THEN    lib$ count "WinLibrary ;


INTERNAL        \ more non-user definitions

\ Special version of WinLibrary that prevents special libraries that are
\ needed at compile time, but not at application runtime from being loaded
\ after the program is turnkeyed.  Specifically provided for WINCON.DLL,
\ but may be used by future DLLs that assist the Win32Forth compiler.

: sys-WinLibrary  { \ lib$ 'name.DLL' -- }  \ usage: WinLibrary user32.dll
        MAXSTRING LocalAlloc: lib$
        /parse-word count lib$ place
        winlib-link @
        begin   ?dup
        while   dup 2 cells+ count lib$ count compare 0=
                if      drop EXIT       \ already in list, EXIT
                then
                @
        repeat
        in-system? abort" DLLs must be in APPLICATION dictionary
        0 ,                             \ the library handlehandle
        winlib-link link,               \ the link of all libraries
        -1 ,                            \ mark as system library
        lib$ count here over 1+ allot place ;

defer dll-error
defer proc-error

\ if not currently running a TURNKEYed application, then return either
\ the procedure address, or a NULL.
\ If we ARE running a TURNKEYed applicaiton and this is a procedure from
\ a system library, then return TRUE, which means just ignore the fact
\ we can't load the library.

: load-winproc ( proc -- proc-addr | NULL | TRUE )
        sys-free 0=                     \ are we running a turnkeyed application
        if      dup cell+ @                     \ get the library address
                2  cells+ @ -1 =                \ is this a SYSTEM library?
                if      dodefer over cell - !   \ change cfa to deferred word
                        ['] dll-error swap !    \ flag error if called
                        TRUE                    \ return TRUE
                        EXIT                    \ leave this word
                then
        then
        dup cell+ @ load-winlibrary 0=
        if      docall over cell - !            \ change CFA to procedure call
                dup cell+ @ @                   \ library handle
                over 3 cells+ count swap rel>abs
                18 xcall                        \ GetProcAddress
                dup rot !
        else    ['] proc-error swap !            \ call error message at runtime
                FALSE
        then    ;

create proc-name MAXSTRING allot
       proc-name off

: ?proc-error   ( f1 -- )       \ f1=error if true
                if      proc-name count pocket place
                        TRUE abort" Couldn't find procedure"
                then    ;

: find-in-lib   ( proc lib -- proc' )
                dup>r cell - load-winlibrary 0=
                if      r@ cell - @ proc-name count swap rel>abs 18 xcall
                        if      drop r@ cell -
                        then
                then    r>drop ;

: find-winlibrary ( -- proc-adr | 0 )   \ find the window library containing the
                                        \ procedure name string in PROC-NAME
        NULL                                    \ initial NULL library address
        ['] find-in-lib winlib-link do-link ;   \ check the non 'a' version

: "proc ( #parms a1 n1 -- )     \ define a procedure from string a1,n1 with #parms
        proc-name place                         \ get the proc name
        >application            \ make sure we are compiling in application
        proc-name +NULL
        drop                                    \ discard the number of parameters
        winproc-link
        begin   @ ?dup
        while   dup cell+ count                 \ next proc name in list
                proc-name count compare 0=      \ is this name already defined
                if      drop                    \ if so, discard linkage
                        application>            \ re-select system memory
                        EXIT                    \ EXIT definition, all done
                then
        repeat
        docall ,                                \ cfa  of do-procedure
        here 0 ,                                \ the procedure address
        find-winlibrary dup 0= ?proc-error ,    \ compile the library address \SMuB
        winproc-link link,                      \ linked list of all procs

        proc-name count here over 1+ allot place \ lay in the procedure name
        align
        load-winproc 0=
        if      TRUE ?proc-error
        then
        application> ;                          \ re-select system memory

EXTERNAL        \ more user accessible definitions

: proc          ( #params -<name>- )   \ #arguments proc MessageBeep
                /parse-word count "proc ;

INTERNAL        \ more non-user definitions

\ -------------------- Calling Procedures --------------------

\ Look for proc name.  If not found, look for name with A appended.
: "find-proc    ( a1 n1 -- cfa-of-proc TRUE | FALSE )
                proc-name place
                winproc-link
                begin   @ ?dup
                while   dup cell+ count proc-name count compare 0=
                        if      3 cells -       \ CFA of procedure
                                TRUE
                                EXIT            \ leave with the CFA
                        then
                repeat  FALSE ;

: res-call      ( -- )          \ resolver for forward referenced procedure calls
                later" 2>r                      \ get the proc name to resolve
                    -1 2r@ "proc                \ define procedure if needed
                       2r> "find-proc 0= ?proc-error
                                                \ get the unresolved address
                later# ! ;                      \ resolve forward reference

EXTERNAL        \ more user accessible definitions

: call          ( [args..] -<proc>- result )    \ compile or execute a windows
                                                \ procedure
                bl word count "find-proc        \ try to find the call
                if                              \ if we found it, then
                        state @                 \ if we are compiling
                        if      ,               \ then compile the call
                        else    execute         \ if interpreting, execute it
                        then
                else                            \ if we didn't find the call
                        state @                 \ and we are compiling, then
                        if                      \ setup for later resolution
                                                \ after definition is finished
                                here #later     \ push address to resolve
                                0 ,             \ place holder for procedure
                                                \ push the name to resolve
                                pocket count "later
                                                \ push the resolver function cfa
                                ['] res-call cfa-later
                        else                    \ otherwise, define the call
                                -1 pocket count "proc
                                                \ look it up
                                pocket count "find-proc 0= ?proc-error
                                execute          \ and then just doit
                        then
                then    ; IMMEDIATE

\ Add FUNC: link SwiftForth, so we can be compatible with their Call mechanism

: FUNC:         ( [args..] -<proc>- result )    \ compile or execute a windows
                                                \ procedure
                0 proc                          \ define the procedure call
                proc-name count "find-proc 0= abort" Can't find procedure"
                proc-name count "header DOVAR , \ CREATE from the name
                IMMEDIATE
                ,                               \ lay in the proc XT
                DOES> @ 
                        state @                 \ if we are compiling
                        if      ,               \ then compile the call
                        else    execute         \ if interpreting, execute it
                        then    ;

: DLL:          { \ lib$ 'name.DLL' -- }  \ usage: DLL: user32<.dll>
                MAXSTRING LocalAlloc: lib$
                /parse-word count lib$ place
                lib$ count + 4 - 4 s" .DLL" caps-compare
                IF      s" .DLL" lib$ +place    \ if no .DLL file extension, add it
                THEN    lib$ count "WinLibrary ;


: .libs         ( -- )
                tab-size >r
                16 to tab-size
                cr
                winlib-link
                begin   @ dup
                        nuf? 0= and
                while   dup 2 cells+ count type tab tab-size ?cr
                repeat  drop cr r> to tab-size ;

: .procs        ( -- )
                tab-size >r
                30 to tab-size
                cr
                winproc-link
                begin   @ dup
                        nuf? 0= and
                while   dup cell+ count type tab tab-size ?cr
                repeat  drop cr r> to tab-size ;

INTERNAL        \ more non-user definitions

fload iid.f

\ -------------------- Load Standard Libraries --------------------

WinLibrary KERNEL32.DLL
WinLibrary GDI32.DLL
WinLibrary USER32.DLL
WinLibrary ADVAPI32.DLL
WinLibrary SHELL32.DLL

\ Don't load the OLE libraries for now

\ WinLibrary OLE32.DLL
\ WinLibrary OLEAUT32.DLL

\ Linkage for Windows Constant server through the number chain

sys-WinLibrary WINCON.DLL       \ Windows Constant server

create WinConPtr                        \ Pointer to FindWin32Constant
        3 proc FindWin32Constant        \ define this entry function first
        3 proc EnumWin32Constants

: wincon-number? ( a1 n1 f1 -- d1 TRUE | a1 n1 FALSE )
                { \ con$ WinVal -- }
                MAXSTRING LocalAlloc: con$      \ allocate a buffer the con name
                WinConPtr cell+ @ 0= ?EXIT      \ Leave if 'FindWin32Constant'
                                                \ hasn't already been resolved
                dup ?EXIT drop
                2dup con$ place                 \ lay string in buffer
                2dup swap rel>abs
                &LOCAL WinVal rel>abs -rot      \ under adr & len
                Call FindWin32Constant
                IF      2drop
                        WinVal 0 TRUE           \ return constant, zero and TRUE
                ELSE    s" A" con$ +place       \ append an 'A'
                        con$ count swap rel>abs
                        &LOCAL WinVal rel>abs -rot      \ under adr & len
                        Call FindWin32Constant
                        IF      2drop
                                WinVal 0 TRUE
                        ELSE    FALSE
                        THEN
                THEN    ;

number?-chain chain-add wincon-number?          \ windows constant server

: 0winproc      ( proc-addr -- )
                2 cells - off ;         \ clear procedure address

: 0winlib       ( lib-addr -- )
                cell - off ;            \ clear the library handle

create errbuf MAXSTRING cell+ allot
       errbuf MAXSTRING cell+ erase

: get-winproc   ( proc-addr -- )        \ load one procedure address
                dup>r
                2 cells - load-winproc 0=
                dup
                if      r@ cell+ count errbuf +place
                        s"  "          errbuf +place
                then    or
                r>drop ;

: init-procedures ( -- )  \ initialize all procedure libraries
                ['] 0winproc    winproc-link do-link    \ clear the procedures
                ['] 0winlib      winlib-link do-link    \ and libraries
                errbuf MAXSTRING cell+ erase            \ clear not found buffer
                FALSE                                   \ initially all OK
                ['] get-winproc winproc-link do-link    \ -- f1=FALSE=All loaded OK
                ignore-missing-procs? 0= and            \ we may want to ignore this warning
                if      MB_OK MB_ICONSTOP or MB_TASKMODAL or
                        z" Missing Procedures!" rel>abs
                        errbuf 1+ rel>abs
                        NULL Call MessageBox  drop
                then    ;

initialization-chain chain-add init-procedures

: _DefaultWindowProc ( hwnd msg wparam lparam -- res )
                swap 2swap swap Call DefWindowProc ;

' _DefaultWindowProc is DefaultWindowProc

EXTERNAL

\ April 23rd, 1999 - 14:05 tjz
\ not updated like WINCON-NUMBER? to succeed on ANSI versions, because
\ this only needs to detect WM_ constants for OOP.

: IsWinConstant ( str -- str FALSE | val TRUE )
                { \ WinVal -- }
                WinConPtr cell+ @ 0=            \ Leave if 'FindWin32Constant'
                if      FALSE EXIT              \ hasn't already been resolved
                then
                &LOCAL WinVal rel>abs over
                count swap rel>abs Call FindWin32Constant
                if      drop
                        WinVal TRUE
                else    FALSE
                then    ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Operating System Checking
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ WindowsNT has it's own power bit

: NT?           ( -- f1 )       \ return true if we are running on NT
                Call GetVersion 0x80000000 and 0= ;

\ Greater than version 4.00 and less than version 4.10 is Windows95

: Win95?        ( -- f1 )       \ return true if we are running Windows 95
                Call GetVersion dup 0x7FFFFFFF and swap 0x80000000 and
                if      word-split drop 0x100 /mod swap 4 < 0=
                        swap 10 < and
                else    drop false
                then    ;

\ Greater than version 4.09 is Windows98

: Win98?        ( -- f1 )       \ return true if we are running Windows 95
                Call GetVersion dup 0x7FFFFFFF and swap 0x80000000 and
                if      word-split drop 0x100 /mod swap 4 < 0=
                        swap 10 < 0= and
                else    drop false
                then    ;

\ less than version 4.00 is Win32s

: Win32s?       ( -- f1 )       \ return true if we are running Win32s
                Call GetVersion dup 0x7FFFFFFF and swap 0x80000000 and
                if      word-split drop 0x100 mod 4 <
                else    drop false
                then    ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Minimal Error Box Functon
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: ErrorBox      ( a1 n1 -- )
                temp$ place
                temp$ +NULL
                MB_TASKMODAL
                z" Application Error" rel>abs
                temp$ 1+ rel>abs
                NULL Call MessageBox drop ;

: _dll-error    ( -- )
                MB_TASKMODAL
                z" Application Error"                   rel>abs
                z" Attempted to use a System DLL Call"  rel>abs
                NULL Call MessageBox drop
                bye ;

' _dll-error is dll-error

: _proc-error   ( -- )
                MB_TASKMODAL
                z" Application Error"                   rel>abs
                z" Attempted to use an unavailable System Call"  rel>abs
                NULL Call MessageBox drop
                bye ;

' _proc-error is proc-error

MODULE          \ finishup the module

